@@ -1,7 +1,12 @@
+/*
+the header is necessary to enable some macros
+#include <openssl/ssl.h>
+*/
+
 /*////////////////////////////////////////////////////////////////////////
 Copyright (c) 1998-2000 Electrotechnical Laboratry (ETL), AIST, MITI
 Copyright (c) 1998-2000 Yutaka Sato
-Copyright (c) 2001-2002 National Institute of Advanced Industrial Science and Technology (AIST)
+Copyright (c) 2001-2005 National Institute of Advanced Industrial Science and Technology (AIST)
 
 Permission to use, copy, modify, and distribute this material for any
 purpose and without fee is hereby granted, provided that the above
@@ -70,10 +75,12 @@ History:
 #include <stdio.h>
 #include <stdlib.h>
 #include "ystring.h"
-#include "randtext.c"
 
-int (*DELEGATE_MAIN)(int ac,const char *av[]);
-void (*DELEGATE_TERMINATE)();
+#ifdef _MSC_VER
+#undef ERROR
+#undef X509
+#undef X509_NAME
+#endif
 
 int randstack_call(int strg,iFUNCP func, ...);
 char **move_envarg(int ac,const char *av[],const char **areap,int *lengp,int *sizep);
@@ -86,6 +93,8 @@ int RecvPeek(int sock,void *buf,int len);
 int isWindows();
 int SocketOf(int sock);
 int LIBFILE_IS(PCStr(file),PVStr(xfile));
+FILE *CFI_fopenShared(PCStr(mode));
+const char *CFI_FILTER_ID();
 
 #define LSILENT	-1
 #define LERROR	0
@@ -99,17 +108,23 @@ static int loglevel = LERROR;
 static FILE *stdctl;
 static const char *client_host;
 static int PID;
+static int Builtin;
 static int DOLOG(PCStr(fmt),...)
 {	CStr(xfmt,256);
 	CStr(head,256);
 	VARGS(8,fmt);
 
+	if( Builtin )
+		sprintf(head,"## SSLway");
+	else
 	sprintf(head,"## SSLway[%d](%s)",PID,client_host);
 	sprintf(xfmt,"%%s %s\n",fmt);
 	syslog_ERROR(xfmt,head,VA8);
 	return 0;
 }
 
+#ifndef SSL_FILETYPE_PEM /*{*/
+/*BEGIN_STAB(ssl)*/
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -122,9 +137,6 @@ extern "C" {
 #define SSL_VERIFY_FAIL_IF_NO_PEER_CERT	0x02
 #define SSL_VERIFY_CLIENT_ONCE		0x04
 
-#define X509_V_FLAG_CRL_CHECK		0x04
-#define X509_V_FLAG_CRL_CHECK_ALL	0x08
-
 typedef void SSL_CTX;
 typedef void SSL_METHOD;
 typedef void SSL;
@@ -134,13 +146,10 @@ typedef void X509_STORE;
 typedef void X509_STORE_CTX;
 typedef void BIO;
 typedef void RSA;
+typedef void SSL_SESSION;
+typedef void EVP_PKEY;
 #define BIO_NOCLOSE 0
 
-unsigned long ERR_get_error(void);
-char *ERR_error_string_n(int,char*,int);
-void ERR_print_errors_fp(FILE *fp);
-void RAND_seed(const void *buf,int num);
-
 BIO *BIO_new_fp(FILE *stream, int close_flag);
 int BIO_free(BIO *a);
 X509 *PEM_read_bio_X509(BIO*,...);
@@ -160,6 +169,36 @@ void SSL_load_error_strings(void );
 int  SSL_get_error(SSL *s,int ret_code);
 X509 *SSL_get_peer_certificate(SSL *ssl);
 
+SSL_SESSION *SSL_SESSION_new(void);
+#define SSL_CTX_sess_set_new_cb(ctx,cb)	/* it's a macro */
+/*
+#define SSL_CTRL_GET_SESSION_REUSED	6
+#define SSL_CTRL_SESS_HIT	27
+#define SSL_CTRL_SESS_MISSES	29
+#define SSL_CTX_sess_hits(ctx) \
+	SSL_CTX_ctrl(ctx,SSL_CTRL_SESS_HIT,0,NULL)
+#define SSL_CTX_sess_misses(ctx) \
+	SSL_CTX_ctrl(ctx,SSL_CTRL_SESS_MISSES,0,NULL)
+#define SSL_session_reused(ssl) \
+	SSL_ctrl((ssl),SSL_CTRL_GET_SESSION_REUSED,0,NULL)
+*/
+#define SSL_session_reused(ssl) 0
+int SSL_CTX_add_session(SSL_CTX *ctx, SSL_SESSION *c);
+SSL_SESSION *SSL_get_session(SSL *ssl);
+int SSL_set_session(SSL *ssl,SSL_SESSION *sess);
+int SSL_CTX_use_certificate(SSL_CTX *ctx, X509 *x);
+int SSL_CTX_use_PrivateKey(SSL_CTX *ctx, EVP_PKEY *pkey);
+X509 *SSL_get_certificate(SSL *ssl);
+EVP_PKEY *SSL_get_privatekey(SSL *ssl);
+#define EVP_PKEY_RSA 6
+SSL_SESSION *d2i_SSL_SESSION(SSL_SESSION **a,unsigned char **pp,long length);
+int i2d_SSL_SESSION(SSL_SESSION *in, unsigned char **pp);
+X509 *d2i_X509(X509 **x,unsigned char**in,int len);
+int i2d_X509(X509 *x,unsigned char **pp);
+EVP_PKEY *d2i_PrivateKey(int type,EVP_PKEY **a,unsigned char **pp,long length);
+int i2d_PrivateKey(EVP_PKEY *a,unsigned char **pp);
+
+long SSL_ctrl(SSL *ssl,int cmd, long larg, void *parg);
 long SSL_CTX_ctrl(SSL_CTX *ctx,int cmd, long larg, void *parg);
 int  SSL_CTX_check_private_key(SSL_CTX *ctx);
 X509_STORE *SSL_CTX_get_cert_store(SSL_CTX *);
@@ -189,15 +228,61 @@ const char *X509_verify_cert_error_string(long n);
 X509 *X509_STORE_CTX_get_current_cert(X509_STORE_CTX *ctx);
 int X509_STORE_CTX_get_error(X509_STORE_CTX *ctx);
 int X509_STORE_CTX_get_error_depth(X509_STORE_CTX *ctx);
-void X509_STORE_set_flags(X509_STORE *ctx, long flags);
 X509_NAME *X509_get_subject_name(X509 *a);
 void X509_free(X509 *a);
 
 RSA *RSA_generate_key(int bits, unsigned long e,void (*callback)(int,int,void *),void *cb_arg);
 
+typedef struct {
+	int	ssl_version;
+ unsigned int	key_arg_length;
+ unsigned char	key_arg[8];
+	int	master_key_length;
+ unsigned char	master_key[48];
+ unsigned int	session_id_length;
+ unsigned char	session_id[32];
+} SessionHead;
+
 #ifdef __cplusplus
 }
 #endif
+#endif /*}*/
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+unsigned long ERR_get_error(void);
+char *ERR_error_string_n(int,char*,int);
+void ERR_print_errors_fp(FILE *fp);
+void RAND_seed(const void *buf,int num);
+void X509_STORE_set_flags(X509_STORE *ctx, long flags);/*OPT(0)*/
+#define X509_V_FLAG_CRL_CHECK		0x04
+#define X509_V_FLAG_CRL_CHECK_ALL	0x08
+
+typedef int (*GEN_SESSION_CB)(const SSL *ssl,unsigned char *id,unsigned int *id_len);
+int SSL_CTX_set_generate_session_id(SSL_CTX *ctx, GEN_SESSION_CB cb);/*OPT(0)*/
+#ifdef __cplusplus
+}
+#endif
+/*END_STAB*/
+
+#ifndef ISDLIB
+#include "randtext.c"
+#endif
+
+#define XACC	0
+#define XCON	1
+
+static double Start;
+static double laps[32];
+static char *lapd[32];
+static int lapx;
+static int do_cache = 1;
+#define Lap(msg)  (loglevel<LDEBUG)?0:((lapd[lapx]=msg),(laps[lapx++]=Time()))
+static int nthcall;
+static SSL_CTX *ctx_cache;
+static int ctx_filter_id;
+
 
 static SSL_CTX *ssl_new(int serv)
 {	SSL_CTX *ctx;
@@ -281,35 +366,192 @@ static void eRR_print_errors_fp(FILE *fp)
 		ERR_print_errors_fp(fp);
 	}
 }
+#undef	ERR_print_errors_fp
 #define	ERR_print_errors_fp	eRR_print_errors_fp
 
+typedef struct {
+ unsigned int	s_leng;
+ unsigned int	s_what;
+ unsigned int	s_date;
+ unsigned char	s_buff[256-sizeof(int)*3];
+} Session;
+static Session sess_cache[8];
+static int sess_cached;
+static int sess_hits;
+
+static void loadSessions(SSL_CTX *ctx,SSL *ssl){
+	FILE *fp;
+	SSL_SESSION *sess;
+	unsigned char sbuf[10*1024];
+	unsigned char *sb;
+	SSL_SESSION *s;
+	int len;
+	int rcc;
+	double start = Time();
+	int nsess;
+	int si;
+	Session *sp;
+
+	if( do_cache == 0 )
+		return;
+
+	fp = CFI_fopenShared("r");
+	if( fp ){
+		int size;
+		fseek(fp,0x4000,0);
+		size = -1;
+		rcc = fread(&size,1,sizeof(size),fp);
+		if( rcc != sizeof(size) || size<=0 || sizeof(sess_cache)<size ){
+			if( rcc != 0 )
+				ERROR("session cache size error: %d",size);
+			return;
+		}
+		rcc = fread(sess_cache,1,size,fp);
+		fclose(fp);
+		if( sizeof(Session) <= rcc )
+		if( (rcc/sizeof(Session))*sizeof(Session) == rcc ){
+			sess_cached = nsess = rcc / sizeof(Session);
+			for( si = 0; si < nsess; si++ ){
+				sp = &sess_cache[si];
+				s = SSL_SESSION_new();
+				sb = sp->s_buff;
+				len = sp->s_leng;
+				if( sp->s_what == XCON ){
+					/*
+					int code;
+					bcopy(sb,s,len);
+					code = SSL_set_session(ssl,s);
+					*/
+					break;
+				}else
+				if( sess = d2i_SSL_SESSION(&s,&sb,len) ){
+					SSL_CTX_add_session(ctx,sess);
+				}
+			}
+			Lap("loadSession OK");
+		}
+	}
+}
+static void saveSessions(SSL_CTX *ctx,SSL *ssl,int what){
+	FILE *fp;
+	SSL_SESSION *sess;
+	unsigned char sbuf[10*1024];
+	unsigned char *sb;
+	int len;
+	double start;
+	int si;
+	int wcc;
+	int nsess;
+	int nsi;
+	Session *sp;
+	unsigned int oldest;
+	int oi;
+	SessionHead *shp;
+
+	if( do_cache == 0 )
+		return;
+
+	start = Time();
+	sess = SSL_get_session(ssl);
+
+	if( what == XCON ){
+		shp = (SessionHead*)sess;
+		ERROR("SESSION leng=%d",shp->session_id_length);
+ERROR(">>>>>>>>>>>> SESSION leng=%d / %d",shp->session_id_length,
+sizeof(SessionHead));
+		len = sizeof(SessionHead);
+		bcopy(shp,sbuf,len);
+	}else{
+		sb = sbuf;
+		len = i2d_SSL_SESSION(sess,&sb);
+	}
+	if( 128-8 < len ){
+		ERROR("Session DER Length: %d",len);
+		return;
+	}
+
+	oi = 0;
+	oldest = 0xFFFFFFFF;
+
+	if( sess_cached ){
+		for( si = 0; si < sess_cached; si++ ){
+			if( bcmp(sess_cache[si].s_buff,sbuf,len) == 0 ){
+				sess_hits++;
+				return;
+			}
+			if( sess_cache[si].s_date < oldest ){
+				oldest = sess_cache[si].s_date;
+				oi = si;
+			}
+		}
+	}
+	nsess = sess_cached;
+	if( nsess == elnumof(sess_cache) ){
+		/* should search last used, least recently hit ... */
+		nsi = oi;
+	}else{
+		nsi = nsess++;
+	}
+	sp = &sess_cache[nsi];
+	sp->s_leng = len;
+	sp->s_what = what;
+	sp->s_date = time(NULL);
+	bcopy(sbuf,sp->s_buff,len);
+
+	fp = CFI_fopenShared("r+");
+	if( fp ){
+		int size = nsess * sizeof(Session);
+		fseek(fp,0x4000,0);
+		fwrite(&size,1,sizeof(size),fp);
+		wcc = fwrite(sess_cache,1,size,fp);
+		fclose(fp);
+		Lap("saveSession OK");
+	}
+}
+
 static SSL *ssl_conn(SSL_CTX *ctx,int confd)
 {	SSL *conSSL;
 
+	Lap("ssl_conn() start");
 	conSSL = SSL_new(ctx);
+/*
+	loadSessions(ctx,conSSL);
+*/
 	SSL_set_connect_state(conSSL);
 	SSL_set_fd(conSSL,SocketOf(confd));
+	Lap("before connect");
 	if( SSL_connect(conSSL) < 0 ){
 		ERROR("connect failed");
 		ERR_print_errors_fp(stderr);
 		return NULL;
 	}else{
+		Lap("after connect");
+/*
+		saveSessions(ctx,conSSL,XCON);
+*/
 		TRACE("connected");
 		return conSSL;
 	}
 }
+
 static SSL *ssl_acc(SSL_CTX *ctx,int accfd)
 {	SSL *accSSL;
 
+	Lap("ssl_acc() start");
+	loadSessions(ctx,NULL);
 	accSSL = SSL_new(ctx);
 	SSL_set_accept_state(accSSL);
 	SSL_set_fd(accSSL,SocketOf(accfd));
+	SSL_set_fd(accSSL,SocketOf(accfd));
+	Lap("before accept");
 	if( SSL_accept(accSSL) < 0 ){
 		ERROR("accept failed");
 		ERR_print_errors_fp(stderr);
 		ssl_printf(accSSL,0,"SSLway: accept failed\n");
 		return NULL;
 	}else{
+		Lap("after accept");
+		saveSessions(ctx,accSSL,XACC);
 		TRACE("accepted");
 		return accSSL;
 	}
@@ -366,9 +608,6 @@ static int   verify_depth = -1;
 static int   do_showCERT = 0;
 static const char *cipher_list = NULL;
 
-#define XACC	0
-#define XCON	1
-
 #define sv_Cert		sslctx[XACC].x_certkey
 #define sv_Ncert	sslctx[XACC].x_certkey.v_Ncert
 #define sv_Nkey		sslctx[XACC].x_certkey.v_Nkey
@@ -449,6 +688,101 @@ static void scanpass(PCStr(arg))
 	}
 }
 
+static int saveContext(SSL_CTX *ctx,SSL *ssl,int ac,char *av[]){
+	X509 *cert;
+	EVP_PKEY *ekey;
+	unsigned char tmp[10*1024];
+	unsigned char *pp;
+	int len;
+	FILE *fp;
+
+	if( do_cache == 0 )
+		return -1;
+
+	if( cert = SSL_get_certificate(ssl) ){
+		if( (ekey = SSL_get_privatekey(ssl)) ){
+			if( (cert = SSL_get_certificate(ssl)) ){
+				if( (fp = CFI_fopenShared("r+")) ){
+					fprintf(fp,"%s\n",CFI_FILTER_ID());
+
+					pp = tmp;
+					len = i2d_X509(cert,&pp);
+					fwrite(&len,1,sizeof(len),fp);
+					fwrite(tmp,1,len,fp);
+
+					pp = tmp;
+					len = i2d_PrivateKey(ekey,&pp);
+					fwrite(&len,1,sizeof(len),fp);
+					fwrite(tmp,1,len,fp);
+
+					fclose(fp);
+					Lap("saveContext OK");
+					return 0;
+				}
+			}
+		}
+	}
+	return -1;
+}
+
+static int loadContext(SSL_CTX *ctx,int ac,char *av[]){
+	FILE *fp;
+	CStr(fid,64);
+	char *dp;
+	X509 *cert;
+	EVP_PKEY *pkey;
+	unsigned CStr(buf,4096);
+	unsigned char *pp;
+	int len;
+	double start = Time();
+
+	if( do_cache == 0 )
+		return -1;
+
+	fp = CFI_fopenShared("r");
+	if( fp == NULL )
+		return -1;
+
+	fgets(fid,sizeof(fid),fp);
+	if( dp = strchr(fid,'\n') )
+		*dp = 0;
+	if( strcmp(fid,CFI_FILTER_ID()) != 0 ){
+		fclose(fp);
+		return -1;
+	}
+	len = -1;
+	fread(&len,1,sizeof(len),fp);
+	if( len <= 0 || sizeof(buf) < len ){
+		fclose(fp);
+		return -1;
+	}
+	fread(buf,1,len,fp);
+	pp = (unsigned char*)buf;
+	cert = d2i_X509(NULL,&pp,len);
+
+	len = -1;
+	fread(&len,1,sizeof(len),fp);
+	if( len <= 0 || sizeof(buf) < len ){
+		fclose(fp);
+		return -1;
+	}
+	fread(buf,1,len,fp);
+	pp = (unsigned char*)buf;
+	pkey = d2i_PrivateKey(EVP_PKEY_RSA,NULL,&pp,len);
+
+	fclose(fp);
+
+	if( SSL_CTX_use_certificate(ctx,cert) ){
+		if( SSL_CTX_use_PrivateKey(ctx,pkey) ){
+			if( SSL_CTX_check_private_key(ctx) ){
+				Lap("loadContext OK");
+				return 0;
+			}
+		}
+	}
+	return -1;
+}
+
 static int setcert1(SSL_CTX *ctx,PCStr(certfile),PCStr(keyfile),int clnt)
 {	int code = 0;
 	CStr(cwd,1024);
@@ -463,6 +797,7 @@ static int setcert1(SSL_CTX *ctx,PCStr(certfile),PCStr(keyfile),int clnt)
 	if( LIBFILE_IS(certfile,AVStr(xcertfile)) )
 		certfile = xcertfile;
 
+	pass[0] = 0;
 	if( dp = strrchr(keyfile,'.') ){
 		passfilename(keyfile,AVStr(passfile));
 		if( 0 <= Freadline(passfile,pass,sizeof(pass)) ){
@@ -491,6 +826,80 @@ static int setcert1(SSL_CTX *ctx,PCStr(certfile),PCStr(keyfile),int clnt)
 	}
 	return code;
 }
+
+int Ftruncate(FILE *fp,int offset,int whence);
+static int cert_opts;
+static const char *dflt_cert;
+static const char *dflt_vkey;
+void sslway_dflt_certkey(PCStr(cert),PCStr(vkey)){
+	dflt_cert = cert;
+	dflt_vkey = vkey;
+}
+
+/*
+#define PEM_STRING_X509		"CERTIFICATE"
+#define PEM_STRING_RSA		"RSA PRIVATE KEY"
+#define PEM_read_X509(fp,x,cb,u) (X509 *)PEM_ASN1_read( \
+	(char *(*)())d2i_X509,PEM_STRING_X509,fp,(char **)x,cb,u)
+#define PEM_read_RSAPrivateKey(fp,x,cb,u) (RSA *)PEM_ASN1_read( \
+	(char *(*)())d2i_RSAPrivateKey,PEM_STRING_RSA,fp,(char **)x,cb,u)
+extern "C" {
+RSA *d2i_RSAPrivateKey(RSA **a,unsigned char **pp,long length);
+int SSL_CTX_use_RSAPrivateKey(SSL_CTX *ctx, RSA *rsa);
+char *PEM_ASN1_read(char *(*d2i)(),const char *name,FILE *fp,char **x,pem_password_cb *cb, void *u);
+}
+*/
+int set_dfltcerts(SSL_CTX *ctx){
+	X509 *cert;
+	EVP_PKEY *pkey;
+	FILE *fp;
+	char file[128];
+
+	if( dflt_cert == 0 || dflt_vkey == 0 )
+		return -1;
+/*
+	fp = tmpfile();
+	fputs(dflt_vkey,fp);
+	fflush(fp);
+	fseek(fp,0,0);
+	cert = NULL;
+	PEM_read_X509(fp,&cert,NULL,NULL);
+	SSL_CTX_use_RSAPrivateKey(ctx,pkey);
+
+	fputs(dflt_cert,fp);
+	fflush(fp);
+	Ftruncate(fp,0,1);
+	fseek(fp,0,0);
+	pkey = NULL;
+	PEM_read_RSAPrivateKey(fp,&pkey,NULL,NULL);
+	SSL_CTX_use_certificate(ctx,pkey);
+	fclose(fp);
+*/
+	tmpnam(file);
+	fp = fopen(file,"w+");
+
+	fputs(dflt_vkey,fp);
+	fflush(fp);
+	Ftruncate(fp,0,1);
+	fseek(fp,0,0);
+	SSL_CTX_use_RSAPrivateKey_file(ctx,file,SSL_FILETYPE_PEM);
+
+	fputs(dflt_cert,fp);
+	fflush(fp);
+	Ftruncate(fp,0,1);
+	fseek(fp,0,0);
+	SSL_CTX_use_certificate_file(ctx,file,SSL_FILETYPE_PEM);
+
+	fclose(fp);
+	unlink(file);
+
+	if( SSL_CTX_check_private_key(ctx) ){
+		ERROR("---- Using Default Certificate");
+		return 0;
+	}else{
+		return -1;
+	}
+}
 static int setcerts(SSL_CTX *ctx,CertKeyV *certv,int clnt)
 {	int certx;
 	int code;
@@ -500,13 +909,19 @@ static int setcerts(SSL_CTX *ctx,CertKeyV *certv,int clnt)
 		cert1 = &certv->v_ck[certx];
 		code = setcert1(ctx,cert1->c_cert,cert1->c_key,clnt);
 		if( code != 0 )
+		{
+			if( cert_opts == 0 ){
+				/* if not specified explicitly */
+				return set_dfltcerts(ctx);
+			}
 			return code;
+		}
 	}
 	return 0;
 }
 
-static void *tmprsa_key;
-static void *tmprsa_callback(void *ctx,int exp,int bits)
+static RSA *tmprsa_key;
+static RSA *tmprsa_callback(SSL *ctx,int exp,int bits)
 {
 	if( bits != 512 && bits != 1024 ){
 		bits = 512;
@@ -516,10 +931,10 @@ static void *tmprsa_callback(void *ctx,int exp,int bits)
 	}
 	return tmprsa_key;
 }
-static int verify_callback(int ok,void *ctx)
+static int verify_callback(int ok,X509_STORE_CTX *ctx)
 {	int err,depth;
 	const char *errsym;
-	void *cert;
+	X509 *cert;
 	CStr(subjb,256);
 
 	cert =   X509_STORE_CTX_get_current_cert(ctx);
@@ -567,6 +982,8 @@ static int SSL_rdwr(int wr,SSL *ssl,void *buf,int siz)
 	}
 	return xcc;
 }
+#undef	SSL_read
+#undef	SSL_write
 #define SSL_read(ss,bf,sz)	SSL_rdwr(0,ss,bf,sz)
 #define SSL_write(ss,bf,sz)	SSL_rdwr(1,ss,(char*)bf,sz)
 
@@ -704,6 +1121,15 @@ static int isinSSL(int fd)
 				return 1;
 			}
 		}
+		else
+		if( buf[0] == 22 ){ /* ConentType:handshake */
+			rcc = RecvPeek(fd,buf,sizeof(buf));
+			ERROR("SSL Hello?%d [%X %d %d %d %d]",rcc,buf[0],
+				buf[1],buf[2],buf[3]<<8|buf[4],buf[5]);
+			if( buf[5] == 1 ){
+				return 1;
+			}
+		}
 	}
 	return 0;
 }
@@ -953,7 +1379,7 @@ static int HTTP_CAresp(int fd,PCStr(certfile))
 	fprintf(tc,"\r\n");
 
 	in = BIO_new_fp(cfp,BIO_NOCLOSE);
-	cert = PEM_read_bio_X509(in,NULL,NULL);
+	cert = PEM_read_bio_X509(in,NULL,NULL,NULL);
 	out = BIO_new_fp(tc,BIO_NOCLOSE);
 	i2d_X509_bio(out,cert);
 
@@ -999,11 +1425,14 @@ static void rand_seed()
 {	int seed[8],si;
 
 	seed[0] = Gettimeofday(&seed[1]);
+	RAND_seed(seed,sizeof(int)*2);
+/*
 	seed[2] = getpid();
 	seed[3] = getuid();
 	seed[4] = (int)seed;
 	seed[5] = (int)rand_seed;
 	RAND_seed(seed,sizeof(seed));
+*/
 	for( si = 0; si < 8; si++ )
 		seed[si] = 0;
 }
@@ -1030,20 +1459,39 @@ static int cl_passwd(char buf[],int siz,int vrfy)
 	return _passwd("CLIENT",cl_pass,cl_key,buf,siz,vrfy);
 }
 
+static SSL_SESSION *get_session_cb(SSL *ssl,unsigned char *id,int len,int *copy){
+	DEBUG("--CB-- GET SESSION %d [%2X]",len,id[0]);
+	return 0;
+}
+static int new_session_cb(SSL *ssl,SSL_SESSION *sess){
+	int len;
+	len = i2d_SSL_SESSION(sess,NULL);
+	DEBUG("--CB-- NEW SESSION CREATED %X, len=%d",sess,len);
+	return 0;
+}
+static int gen_session_cb(const SSL *ssl,unsigned char *id,unsigned int *id_len){	int i;
+
+	DEBUG("--CB-- GEN SESSION CB, len=%d",*id_len);
+	return 1;
+/*
+ for(i = 0; i < *id_len; i++)
+ fprintf(stderr," %02X",id[i]);
+ fprintf(stderr,"\n");
+	id[0] = 'X';
+	*id_len = 1;
+	return 1;
+*/
+}
+
 static void put_help()
 {
-	syslog_ERROR("SSLway 2004-12-5 <ysato@delegate.org>\r\n");
+	syslog_ERROR("SSLway 2005-05-03 <ysato@delegate.org>\r\n");
 }
-extern int mainX(int ac,char*av[]);
-extern int RANDSTACK_RANGE;
 
-int main(int ac,char *av[])
-{
-	randtext(-1);
-	RANDSTACK_RANGE = 256;
-	return randstack_call(1,(iFUNCP)mainX,ac,av);
-}
+/*
 int mainX(int ac,char *av[])
+*/
+int sslway_mainX(int ac,char *av[],int client,int server,int bi)
 {	int ai;
 	const char *arg;
 	int accfd,confd;
@@ -1054,11 +1502,44 @@ int mainX(int ac,char *av[])
 	int vflag;
 	int ctrlopt = 0;
 	int vflags = 0;
-	void *store;
+	X509_STORE *store;
 	int nodelay = 1;
+	int fid = -1;
+	int ctx_reuse = 0;
+	int sreused = 0;
 
+	Start = Time();
+	lapx = 0;
+	for( ai = 1; ai < ac; ai++ ){
+		arg = av[ai];
+		if( strcmp(arg,"-help") == 0 || strcmp(arg,"-v") == 0 ){
+			put_help();
+			exit(0);
+		}else 
+		if( strncmp(arg,"-vv",3) == 0 || strncmp(arg,"-vd",3) == 0 ){
+			loglevel = LDEBUG;
+		}else
+		if( strncmp(arg,"-vu",3) == 0 ){
+			loglevel = LTRACE;
+		}else
+		if( strncmp(arg,"-vt",3) == 0 ){
+			loglevel = LERROR;
+		}else
+		if( strncmp(arg,"-vs",3) == 0 ){
+			loglevel = LSILENT;
+		}
+	}
+	Lap("start");
+
+	nthcall++;
+	if( env = getenv("CFI_FILTER_ID") )
+		fid = atoi(env);
+	/*
+	if( !bi )
+	CFI_SHARED_FD or so is necessary
+	*/
 	CFI_init(ac,(const char**)av);
-	av = move_envarg(ac,(const char**)av,NULL,NULL,NULL);
+	Lap("init done");
 
 	if( env = getenv("CFI_TYPE") ){
 		if( strcmp(env,"FCL") == 0 ){
@@ -1088,8 +1569,17 @@ int mainX(int ac,char *av[])
 	if( (client_host = getenv("REMOTE_HOST")) == 0 )
 		client_host = "?";
 
+/*
 	accfd = dup(0);
 	confd = dup(1);
+*/
+	if( 0 <= client ){
+		accfd = dup(client);
+		confd = dup(server);
+	}else{
+		accfd = -1;
+		confd = -1;
+	}
 
 	if( env = getenv("SSL_CIPHER") ) cipher_list = env;
 
@@ -1102,23 +1592,18 @@ int mainX(int ac,char *av[])
 	if( env = getenv("SSL_CLIENT_KEY_FILE" ) ) cl_key  = env;
 	if( env = getenv("SSL_CLIENT_CERT_FILE") ) cl_cert = env;
 
+	Lap("begin args");
 	for( ai = 1; ai < ac; ai++ ){
 		arg = av[ai];
 		if( strcmp(arg,"-help") == 0 || strcmp(arg,"-v") == 0 ){
-			put_help();
-			exit(0);
 		}else 
 		if( strncmp(arg,"-vv",3) == 0 || strncmp(arg,"-vd",3) == 0 ){
-			loglevel = LDEBUG;
 		}else
 		if( strncmp(arg,"-vu",3) == 0 ){
-			loglevel = LTRACE;
 		}else
 		if( strncmp(arg,"-vt",3) == 0 ){
-			loglevel = LERROR;
 		}else
 		if( strncmp(arg,"-vs",3) == 0 ){
-			loglevel = LSILENT;
 		}else
 		if( strncasecmp(arg,"-ss",3) == 0 ){
 			do_accSTLS = do_conSTLS = ST_SSL;
@@ -1198,6 +1683,7 @@ int mainX(int ac,char *av[])
 		}else
 		if( strcmp(arg,"-certkey") == 0
 		 || strcmp(arg,"-cert") == 0 ){
+			cert_opts++;
 			if( ac <= ai + 1 ){
 				ERROR("Usage: %s cert-key-file-name",arg);
 				return -1;
@@ -1239,6 +1725,10 @@ int mainX(int ac,char *av[])
 			ctrlopt = 0x000FFFFFL; /* SSL_OP_ALL */
 		}
 		else
+		if( strcmp(arg,"-nocache") == 0 ){
+			do_cache = 0;
+		}
+		else
 		if( strcmp(arg,"-delay") == 0 ){
 			nodelay = 0;
 		}
@@ -1252,19 +1742,27 @@ int mainX(int ac,char *av[])
 				| X509_V_FLAG_CRL_CHECK_ALL;
 		}
 	}
+	Lap("end args");
 
 	accSSL = NULL;
 	conSSL = NULL;
 
 	if( do_conSSL || do_accSSL )
 	{
+		if( nthcall <= 1 ){
 		rand_seed();
+		Lap("end rand_seed");
+		}
 		TRACE("start");
 	}
 
+	if( accfd < 0 ){
+		/* setting ctx */
+	}else{
 	if( nodelay ){
 		set_nodelay(accfd,1);
 		set_nodelay(confd,1);
+		Lap("nodelay set");
 	}
 
 	fdv[0] = accfd;
@@ -1283,12 +1781,13 @@ int mainX(int ac,char *av[])
 			}
 		}
 	}
-
 	if( do_conSSL && do_conSTLS || do_accSSL && do_accSTLS ){
 		if( starttls(accfd,confd) < 0 )
 			return -1;
 	}
+	}
 
+	Lap("start con/acc");
 	if( do_conSSL ){
 		ctx = ssl_new(0);
 		SSL_CTX_set_default_passwd_cb(ctx,(pem_password_cb*)cl_passwd);
@@ -1306,20 +1805,49 @@ int mainX(int ac,char *av[])
 			X509_STORE_set_flags(store, vflags);
 		}
 
+		if( 0 <= confd ){
 		conSSL = ssl_conn(ctx,confd);
 		if( conSSL == NULL )
 			return -1;
+		}
+		sreused += SSL_session_reused(conSSL);
 	}
 
 	if( do_accSSL ){
+		if( fid == ctx_filter_id && ctx_cache != NULL ){
+			ctx = ctx_cache;
+			TRACE("reuse ctx #%d %X",fid,ctx);
+			ctx_reuse = 1;
+		}else{
+		Lap("before ssl_new");
 		ctx = ssl_new(1);
+		Lap("after ssl_new");
+
+		ctx_filter_id = fid;
+		ctx_cache = ctx;
+		TRACE("new ctx #%d %X",fid,ctx);
+
+/*
+SSL_CTX_sess_set_new_cb(ctx,new_session_cb);
+SSL_CTX_sess_set_get_cb(ctx,get_session_cb);
+SSL_CTX_set_session_id_context(ctx,"sslway",6);
+SSL_CTX_set_generate_session_id(ctx,gen_session_cb);
+*/
+
 		if( ctrlopt )
 			SSL_CTX_ctrl(ctx,32/*SSL_CTRL_OPTIONS*/,ctrlopt,NULL);
 		SSL_CTX_set_default_passwd_cb(ctx,(pem_password_cb*)sv_passwd);
 		if( cipher_list )
 			SSL_CTX_set_cipher_list(ctx,cipher_list);
+
+		Lap("before loadContext");
+		if( loadContext(ctx,ac,av) == 0 ){
+			ctx_reuse = 2;
+		}else{
 		if( setcerts(ctx,&sv_Cert,0) < 0 )
 			return -1;
+		}
+		Lap("after loadContext");
 		SSL_CTX_set_tmp_rsa_callback(ctx,tmprsa_callback);
 
 		if( cl_CAfile || cl_CApath )
@@ -1330,10 +1858,36 @@ int mainX(int ac,char *av[])
 			store = SSL_CTX_get_cert_store(ctx);
 			X509_STORE_set_flags(store, vflags);
 		}
+		}
 
+		if( 0 <= accfd ){
 		accSSL = ssl_acc(ctx,accfd);
 		if( accSSL == NULL )
 			return -1;
+
+		if( ctx_reuse == 0 )
+		if( sess_hits == 0 ){
+			saveContext(ctx,accSSL,ac,av);
+		}
+		}
+	}
+
+	Lap("start relay ...");
+	if( sess_cached || sreused || ctx_reuse )
+ERROR("## %f sescache[%d] HIT=%d sR=%d cR=%d",
+		Time()-Start,sess_cached,sess_hits,sreused,ctx_reuse);
+	else
+	if( loglevel <= LERROR )
+ERROR("## %f connected/accepted",Time()-Start);
+	if( LDEBUG <= loglevel ){
+		int i;
+		for( i = 0; i < lapx; i++ )
+DEBUG("-- %f %s",laps[i]-Start,lapd[i]);
+	}
+
+	if( accfd < 0 ){
+		ERROR("initialized ctx #%d %X %X",fid,conSSL,accSSL);
+		return 0;
 	}
 
 	if( conSSL )
@@ -1347,3 +1901,60 @@ int mainX(int ac,char *av[])
 		TRACE("done");
 	return 0;
 }
+#ifdef ISDLIB /*{*/
+int dl_library(const char *libname,DLMap *dlmap,const char *mode);
+static int with_dl;
+int sslway_dl(){
+	if( with_dl ){
+		if( 0 < with_dl )
+			return 1;
+		else	return 0;
+	}
+#ifdef _MSC_VER
+	dl_library("ssleay32.dll",dlmap_ssl,"");
+	if( dl_library("libeay32.dll",dlmap_ssl,"") == 0 ){
+		with_dl = 1;
+		return 1;
+	}else
+#else
+	if( dl_library("ssl",dlmap_ssl,"") == 0 ){
+		with_dl = 1;
+		return 1;
+	}else
+#endif
+	{
+		with_dl = -1;
+		return 0;
+	}
+}
+int sslway_main(int ac,const char *av[])
+{
+	if( sslway_dl() == 0 ){
+		fprintf(stderr,"Can't link the SSL library.\n");
+		return -1;
+	}
+	Builtin = 1;
+	return sslway_mainX(ac,(char**)av,0,1,0);
+}
+int sslwayFilter(int ac,char *av[],FILE *in,FILE *out,int internal){
+	if( sslway_dl() == 0 ){
+		return 0;
+	}
+	Builtin = 1;
+	if( in == NULL )
+		sslway_mainX(ac,av,-1,-1,1);
+	else	sslway_mainX(ac,av,fileno(in),fileno(out),1);
+	return 1;
+}
+#else /*}{*/
+int (*DELEGATE_MAIN)(int ac,const char *av[]);
+void (*DELEGATE_TERMINATE)();
+extern int RANDSTACK_RANGE;
+int main(int ac,char *av[])
+{
+	randtext(-1);
+	RANDSTACK_RANGE = 256;
+	av = move_envarg(ac,(const char**)av,NULL,NULL,NULL);
+	return randstack_call(1,(iFUNCP)sslway_mainX,ac,av,0,1,0);
+}
+#endif /*}*/
