@@ -4,14 +4,17 @@
 typedef struct {
   unsigned int	s_sp0;
 	char   *s_what;
-	int   (*s_func)();
+	int   (*s_func)(const void*,...);
 	char   *s_av[4];
 	int	s_size;
 	int	s_unit;
 	int	s_count;
 	int	s_trace;
+	int	s_level;
 } AllocaArg;
 
+int alloca_call(AllocaArg *ap);
+
 extern int RAND_TRACE;
 #define STACK1	128
 
