@@ -1,8 +1,8 @@
 #include <stdio.h>
 #include "yalloca.h"
+int porting_dbg(const char *fmt,...);
 
-alloca_call(ap)
-	AllocaArg *ap;
+int alloca_call(AllocaArg *ap)
 {	char buff[STACK1];
 	int size;
 
@@ -18,4 +18,4 @@ alloca_call(ap)
 				ap->s_av[2],ap->s_av[3]);
 	}
 }
-INHERENT_alloca(){ return 0; }
+int INHERENT_alloca(){ return 0; }
