@@ -20,4 +20,5 @@ History:
 	980529	created
 //////////////////////////////////////////////////////////////////////#*/
 #include <stdio.h>
-main(ac,av) char *av[]; { return ccx_main(ac,av); }
+int ccx_main(int ac,const char *av[]);
+int main(int ac,char *av[]){ return ccx_main(ac,(const char**)av); }
