@@ -307,7 +307,7 @@ int Str_sprintf(StrHead *StrH,PCStr(fmt), ...)
 					Xstrcpy(QVStr(dp+leng,StrH->sh_base),"=-=\n");
 					dp += leng + 4;
 				}
-				goto OUT;
+				goto OUTx;
 			}
 			break;
 		  case 'c':
@@ -325,7 +325,7 @@ int Str_sprintf(StrHead *StrH,PCStr(fmt), ...)
 		Xsprintf(QVStr(dp,StrH->sh_base),(char*)fmt1,v1);
 		dp += strlen((char*)dp);
 	}
-OUT:
+OUTx:
 	if( dp <= xp )
 		*dp = 0;
 	fmtOK++;
