@@ -30,6 +30,7 @@ int file_ISSOCK(int fd);
 void msleep(int msec);
 
 int RES_getns1(int nsi,VSAddr *sin);
+int RES_proxy();
 int VSA_comp(VSAddr*,VSAddr*);
 static int getns1(int nid,int nsx,int nsc,VSAddr *nservv);
 
@@ -53,7 +54,7 @@ static const char *SYM_TYPE[] = {
 	"XPTR",		"TYPE=21?",	"TYPE=22?",	"TYPE=23?",
 	"TYPE=24?",	"TYPE=25?",	"TYPE=26?",	"TYPE=27?",
 	"AAAA",		"TYPE=29?",	"TYPE=30?",	"TYPE=31?",
-	"TYPE=32?",	"TYPE=33?",	"TYPE=34?",	"TYPE=35?",
+	"TYPE=32?",	"SRV",		"TYPE=34?",	"TYPE=35?",
 	"TYPE=36?",	"TYPE=37?",	"A6",		"TYPE=39?",
 };
 #define symTYPE(class)  ((0<=class&&class<=39) ? SYM_TYPE[class]:\
@@ -270,6 +271,16 @@ static void hexd(PCStr(fmt),int col,octet *str,int len)
 	}
 	debug(DBG_ALL,"\n");
 }
+
+#define putLongVStr(rp,iv) {\
+	setVStrPtrInc(rp,0xFF & (iv >> 24)); \
+	setVStrPtrInc(rp,0xFF & (iv >> 16)); \
+	setVStrPtrInc(rp,0xFF & (iv >>  8)); \
+	setVStrPtrInc(rp,0xFF & (iv >>  0)); \
+}
+
+static char LastSOAdom[128]; /* enough to hold x.x.x.x...IP6.INT */
+
 static
 char *dumpResourceRecord(int whatx,int rri,PCStr(msg),int size,PCStr(rr))
 {	CStr(name,512);
@@ -280,7 +291,10 @@ char *dumpResourceRecord(int whatx,int rri,PCStr(msg),int size,PCStr(rr))
 	const char *stype;
 	const char *sclass;
 	int nid,rnid;
+	/*
 	CStr(addr,32);
+	*/
+	CStr(addr,64);
 	CStr(rdata,512);
 	refQStr(rp,rdata);
 	int pref;
@@ -300,6 +314,7 @@ char *dumpResourceRecord(int whatx,int rri,PCStr(msg),int size,PCStr(rr))
 	getLong(mp,ttl);
 	getShort(mp,rdlength);
 
+	if( type != TY_AAAA )
 	if( type <= 0 || 16 < type || xclass <= 0 ||  4 < xclass ){
 debug(DBG_ANY,"### %s[%2d]%8d: <%d>%d class=%x type=%x\n",
 	what,rri,ttl,nid,name,xclass,type);
@@ -319,16 +334,35 @@ debug(DBG_ANY,"### %s[%2d]%8d: <%d>%d class=%x type=%x\n",
 	setVStrPtrInc(rp,xclass);
 
 	switch( type ){
+	    case TY_SRV:
+		debug(DBG_FORCE,"SRV dump is not supported\n");
+		break;
+
+	    case TY_AAAA:
+		strcpy(addr,VSA_ltoa((unsigned char*)mp,16,AF_INET6));
+		rnid = DNS_putbyaddr(addr);
+		bcopy(mp,(char*)rp,16);
+		rp += 16;
+		DNS_putattr(nid,typemask(type),ttl,rdata,rp-rdata);
+		debug(DBG_RR,"<%d>%s\n",rnid,addr);
+		break;
+
 	    case TY_A:
+		/*
 		sprintf(addr,"%d.%d.%d.%d",mp[0],mp[1],mp[2],mp[3]);
+		*/
+		strcpy(addr,VSA_ltoa((unsigned char*)mp,4,AF_INET));
 		rnid = DNS_putbyaddr(addr);
 		setVStrPtrInc(rp,mp[0]);
 		setVStrPtrInc(rp,mp[1]);
 		setVStrPtrInc(rp,mp[2]);
 		setVStrPtrInc(rp,mp[3]);
 		DNS_putattr(nid,typemask(type),ttl,rdata,6);
+		debug(DBG_RR,"<%d>%s\n",rnid,addr);
+		/*
 		debug(DBG_RR,"<%d>%d.%d.%d.%d\n",
 			rnid,mp[0],mp[1],mp[2],mp[3]);
+		*/
 
 /*
 may be useful for inverse query, but could be bad for consistency...
@@ -376,7 +410,10 @@ debug(DBG_FORCE,"%s <%d>%d.%d.%d.%d\n",name,rnid,mp[0],mp[1],mp[2],mp[3]);
 CStr(mname,512);
 CStr(rname,512);
 int serial,refresh,retry,expire,minimum;
+		int mnid,rnid;
+		mp =
 		scanName(msg,mp,AVStr(mname),sizeof(mname),0);
+		mp =
 		scanName(msg,mp,AVStr(rname),sizeof(rname),0);
 		getLong(mp,serial);
 		getLong(mp,refresh);
@@ -385,6 +422,18 @@ int serial,refresh,retry,expire,minimum;
 		getLong(mp,minimum);
 		debug(DBG_RR,"%s %s %d %d %d %d %d\n",mname,rname,
 			serial,refresh,retry,expire,minimum);
+
+		mnid = DNS_putbyname(mname);
+		rnid = DNS_putbyname(rname);
+		putLongVStr(rp,mnid);
+		putLongVStr(rp,rnid);
+		putLongVStr(rp,serial);
+		putLongVStr(rp,refresh);
+		putLongVStr(rp,retry);
+		putLongVStr(rp,expire);
+		putLongVStr(rp,minimum);
+		DNS_putattr(nid,typemask(type),ttl,rdata,rp-rdata);
+		Xstrcpy(FVStr(LastSOAdom),name);
 }
 		break;
 	    default:
@@ -406,17 +455,33 @@ static int makeMysock()
 {	int len;
 	int rcode;
 	int mysock;
+	int af;
+	const char *me;
 
 	if( Me == 0 )
 		Me = NewStruct(VSAddr);
 
+	/*
 	mysock = socket(AF_INET,SOCK_DGRAM,0);
+	*/
+	af = AF_INET;
+/*
+af = AF_INET6;
+*/
+	mysock = socket(af,SOCK_DGRAM,0);
 	if( mysock < 0 ){
 		debug(DBG_FORCE,"RESOLVY cannot get socket() [errno=%d]\n",
 			errno);
 		return mysock;
 	}
+	/*
 	len = VSA_atosa(Me,0,"0.0.0.0");
+	*/
+	me = "0.0.0.0";
+/*
+me = "::";
+*/
+	len = VSA_atosa(Me,0,me);
 	rcode = bind(mysock,(SAP)Me,len);
 	if( rcode < 0 ){
 		debug(DBG_FORCE,"cannot bind() [errno=%d]\n",errno);
@@ -485,6 +550,19 @@ static int lookupCache(int nid,int qtype,int rri,int rrc,char *rrv[])
 				continue;
 			}
 
+			if( type == TY_SOA ){
+				DNS_nodename(ival,AVStr(dname));
+				debug(DBG_CACHE,"[%d] SOA %s\n",ai,dname);
+				rrv[rri++] = av[ai];
+			}else
+			if( type == TY_SRV ){
+				debug(DBG_FORCE,"SRV lookup is not supprted\n");
+			}else
+			if( type == TY_AAAA ){
+				debug(DBG_CACHE,"[%d] A %s\n",
+					ai,VSA_ltoa(a1,16,AF_INET6));
+				rrv[rri++] = av[ai];
+			}else
 			if( type == TY_A ){
 				debug(DBG_CACHE,"[%d] A %d.%d.%d.%d\n",
 					ai,a1[0],a1[1],a1[2],a1[3]);
@@ -986,6 +1064,7 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 	int rri,rro;
 	int ival;
 	CStr(dname,512);
+	CStr(CNAME,512);
 	refQStr(rrp,rrb);
 	int start;
 	int timeout;
@@ -993,6 +1072,11 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 	int mx,nmx,mxi,rmx;
 	char *mxv[64]; /**/
 
+	if( strncasecmp(name,"_srv.",5) == 0 ){
+		name += 5;
+		qtype = TY_SRV;
+	}
+
 	mx = qtype == TY_A && strncmp(name,"-MX.",4) == 0;
 	if( mx ){
 		name += 4;
@@ -1030,7 +1114,10 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 	}
 
 	start = time(0);
+	/*
 	if( qtype == TY_A )
+	*/
+	if( qtype == TY_A || qtype == TY_AAAA )
 		timeout = RSLV_TIMEOUT;
 	else	timeout = RSLV_INV_TIMEOUT;
 
@@ -1042,12 +1129,20 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 	}
 
 	rro = 0;
+	CNAME[0] = 0;
 	for( rri = 0; rri < rrc; rri++ ){
 		type = rrv[rri][0];
 		rp = rrv[rri] = &rrv[rri][2];
 		rrv[rro] = rrv[rri];
 
 		switch( type ){
+		    case TY_AAAA:
+			rrv[rro] = (char*)rrp;
+			Bcopy(rp,rrp,16);
+			rrp += 16;
+			rro++;
+			break;
+
 		    case TY_A:
 			/*printf("[%d] %d.%d.%d.%d\n",
 				rri,rp[0],rp[1],rp[2],rp[3]);*/
@@ -1066,6 +1161,7 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 			getLong(rp,ival);
 			DNS_nodename(ival,AVStr(dname));
 			if( type == TY_CNAME ){
+				strcpy(CNAME,dname);
 				if( cname != 0 )
 					strcpy(cname,dname);
 				if( qtype != TY_CNAME && qtype != TY_PTR )
@@ -1082,6 +1178,15 @@ int gethostbynameaddr_dns(PCStr(name),int qtype,int rrc,char *rrv[],PVStr(rrb),P
 	rrv[rro+1] = (char*)rrp; /* return value to MX search */
 
 	debug(DBG_NS,"%d seconds, %d(%d) records\n",time(0)-start,rro,rrc);
+
+	if( rro == 0 && 0 < rri )
+	if( qtype == TY_A || qtype == TY_AAAA )
+	if( cname != 0 && CNAME[0] && strcasecmp(name,CNAME) != 0 )
+	{
+		debug(DBG_ANY,"search A/AAAA of '%s' for '%s'\n",CNAME,name);
+		rro =
+		gethostbynameaddr_dns(CNAME,qtype,rrc,rrv,AVStr(rrb),VStrNULL);
+	}
 	return rro;
 }
 
@@ -1096,12 +1201,29 @@ int   DNS_RETRY;
 int   DNS_EXPIRE;
 int   DNS_MINTTL;
 
+static struct hostent *gethostbyNX(PCStr(name),PCStr(af),PCStr(domain));
 static struct hostent *gethostbyN(PCStr(name))
+{
+	return gethostbyNX(name,0,0);
+}
+static struct hostent *gethostbyNX(PCStr(name),PCStr(af),PCStr(domain))
 {	struct hostent *ht;
 	CStr(xname,512);
 	const char *dp;
+	const char *saf;
+	const char *sdomain;
+
+	if( af ){
+		saf = RES_AF;
+		RES_AF = (char*)af;
+	}
+	if( domain ){
+		sdomain = DNS_DOMAIN;
+		DNS_DOMAIN = domain;
+	}
 
 	ht = NULL;
+	if( DNS_DOMAIN[0] )
 	if( dp = strcasestr(name,DNS_DOMAIN) )
 	if( dp[strlen(DNS_DOMAIN)] == 0 ){
 		strcpy(xname,name);
@@ -1113,10 +1235,19 @@ static struct hostent *gethostbyN(PCStr(name))
 	}
 	if( ht == NULL )
 		ht = _GETHOSTBYNAME(name);
+
+	if( DNS_DOMAIN[0] )
 	if( ht == NULL ){
 		sprintf(xname,"%s.%s",name,DNS_DOMAIN);
 		ht = _GETHOSTBYNAME(xname);
 	}
+
+	if( af ){
+		RES_AF = (char*)saf;
+	}
+	if( domain ){
+		DNS_DOMAIN = (char*)sdomain;
+	}
 	return ht;
 }
 static
@@ -1141,10 +1272,16 @@ static int putRR_A(PVStr(rrb),char **rrpp,PCStr(name))
 
 	rrp = *rrpp;
 
+/*
 	ht = gethostbyN(name);
+*/
+	ht = gethostbyNX(name,"4","");
 	if( ht == NULL )
 		return 0;
 
+	if( ht->h_addrtype == AF_INET6 )
+		return 0;
+
 	anc = 0;
 	rttl = DNS_MINTTL;
 	rrdlength = ht->h_length;
@@ -1157,6 +1294,36 @@ static int putRR_A(PVStr(rrb),char **rrpp,PCStr(name))
 	*rrpp = (char*)rrp;
 	return anc;
 }
+static int putRR_AAAA(PVStr(rrb),char **rrpp,PCStr(name))
+{	refQStr(rrp,rrb);
+	struct hostent *ht;
+	int anc,rttl,rrdlength;
+	int ai;
+	const char *baddr;
+	CStr(xname,256);
+
+	rrp = *rrpp;
+
+	sprintf(xname,"-AAAA.%s",name);
+/*
+	ht = gethostbyN(xname);
+*/
+	ht = gethostbyNX(xname,"6","");
+	if( ht == NULL )
+		return 0;
+
+	anc = 0;
+	rttl = DNS_MINTTL;
+	rrdlength = ht->h_length;
+	for( ai = 0; ht->h_addr_list[ai] != 0; ai++ ){
+		baddr = ht->h_addr_list[ai];
+		rrp = putRR(AVStr(rrp),name,TY_AAAA,rttl,rrdlength,baddr);
+		anc++;
+	}
+
+	*rrpp = (char*)rrp;
+	return anc;
+}
 static
 int putRR_name(char **rrpp,xPVStr(rrp),struct hostent *ht,int type,PCStr(name))
 {	int anc;
@@ -1206,16 +1373,29 @@ int putRR_name(char **rrpp,xPVStr(rrp),struct hostent *ht,int type,PCStr(name))
 	*rrpp = (char*)rrp;
 	return anc;
 }
+static int relay_cachedRRs(PVStr(rrb),char **rrrp,PCStr(name),int qtype);
 static int putRR_SOA(PVStr(rrb),char **rrpp,PCStr(name))
 {	refQStr(rrp,rrb);
 	CStr(rbuff,512);
 	refQStr(np,rbuff);
 	int anc,rttl,rrdlength;
 	struct hostent *ht;
+	int rrc;
+	char *rrv[4];
 
 	rrp = *rrpp;
 
+	if( strtailstr(name,DNS_DOMAIN) == NULL ){
+		rrc = getRRbynameaddr(RSLV_TIMEOUT,name,TY_SOA,4,rrv);
+		anc = relay_cachedRRs(AVStr(rrb),rrpp,name,TY_SOA);
+		if( 0 < anc )
+			return anc;
+	}
+
+/*
 	ht = gethostbyN(name);
+*/
+	ht = gethostbyNX(name,"4","");
 	if( ht == NULL )
 		return 0;
 
@@ -1267,6 +1447,24 @@ static int relay_cachedRR(PVStr(rrb),char **rrrp,PCStr(name),int qtype)
 		getLong(ap,ival);
 		np = rbuff;
 
+		if( type == TY_SOA ){
+			DNS_nodename(ival,AVStr(dname));
+			np = putName(AVStr(np),dname);
+			getLong(ap,ival);
+			DNS_nodename(ival,AVStr(dname));
+			np = putName(AVStr(np),dname);
+			getLong(ap,ival); putLongVStr(np,ival);
+			getLong(ap,ival); putLongVStr(np,ival);
+			getLong(ap,ival); putLongVStr(np,ival);
+			getLong(ap,ival); putLongVStr(np,ival);
+			getLong(ap,ival); putLongVStr(np,ival);
+			rttl = DNS_MINTTL;
+			rrdlength = np - rbuff;
+			rrp = putRR(AVStr(rrp),name,type,rttl,rrdlength,rbuff);
+			anc++;
+			continue;
+		}
+		/* MX */
 		getShort(ap,pri); /* MX preference */
 		putShort((char*)np,pri);
 
@@ -1280,6 +1478,45 @@ static int relay_cachedRR(PVStr(rrb),char **rrrp,PCStr(name),int qtype)
 	*rrrp = (char*)rrp;
 	return anc;
 }
+static int relay_cachedRRs(PVStr(rrb),char **rrrp,PCStr(name),int qtype)
+{	int anc;
+
+	anc = relay_cachedRR(AVStr(rrb),rrrp,name,qtype);
+	if( anc == 0 ){
+		/* _nicname._tcp.domain _service._proto.domain RFC2782
+		 * the RRs of it belongs to "domain", so find the domain
+		 */
+		if( *name == '_' ){
+			const char *hp;
+			const char *dp;
+			for( hp = name; *hp == '_'; ){
+				if( dp = strchr(hp,'.') )
+					hp = dp + 1;
+				else	break;
+			}
+			anc = relay_cachedRR(AVStr(rrb),rrrp,hp,qtype);
+		}
+		if( anc == 0 )
+		if( LastSOAdom[0] && strtailstr(name,LastSOAdom) ){
+			/* host.domain
+			 * SOA record of the domain belongs to "domain", so...
+			 */
+			anc = relay_cachedRR(AVStr(rrb),rrrp,LastSOAdom,qtype);
+		}
+	}
+	return anc;
+}
+static int putRR_SRV(PVStr(rrb),char **rrpp,PCStr(name))
+{	refQStr(rrp,rrb);
+	int anc;
+	int rrc;
+	char *rrv[4];
+
+	rrp = *rrpp;
+	rrc = getRRbynameaddr(RSLV_TIMEOUT,name,TY_SRV,elnumof(rrv),rrv);
+	anc = relay_cachedRRs(AVStr(rrb),rrpp,name,TY_SRV);
+	return anc;
+}
 
 /*
  * to be used in hosts or NIS as
@@ -1322,7 +1559,9 @@ static int putRR_PTR(PVStr(rrb),char **rrpp,PCStr(name))
 	rrp = *rrpp;
 
 	if( strcasestr(name,REVERSE_DOM) == 0 )
+	if( strcasestr(name,".IP6.INT") == 0 )
 		return 0;
+
 	if( VSA_dnstosa(&sab,0,name) <= 0 )
 		return 0;
 	bleng = VSA_decomp(&sab,&baddr,&btype,NULL);
@@ -1338,6 +1577,7 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 	refQStr(rp,reply);
 	Header Head,RHead;
 	int QM,RM,qdc,anc,nsc,arc;
+	int nauth;
 	CStr(name,512);
 	CStr(RRbuf,2048);
 	refQStr(rrp,RRbuf);
@@ -1364,9 +1604,14 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 	rrp = RRbuf;
 
 	anc = 0;
+	nauth = 0;
 	switch( qtype ){
+	    case TY_SRV:anc += putRR_SRV(AVStr(RRbuf),(char**)&rrp,name); break;
+/*
 	    case TY_SOA:anc += putRR_SOA(AVStr(RRbuf),(char**)&rrp,name); break;
+*/
 	    case TY_A:  anc += putRR_A(AVStr(RRbuf),(char**)&rrp,name);   break;
+	    case TY_AAAA:anc +=putRR_AAAA(AVStr(RRbuf),(char**)&rrp,name);break;
 	    case TY_PTR:anc += putRR_PTR(AVStr(RRbuf),(char**)&rrp,name); break;
 
 	    case TY_NS:
@@ -1378,14 +1623,22 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 
 	    case TY_QALL:
 			anc += putRR_A(AVStr(RRbuf),(char**)&rrp,name);
+			anc += putRR_AAAA(AVStr(RRbuf),(char**)&rrp,name);
 			anc += putRR_MX(AVStr(RRbuf),(char**)&rrp,name);
 			anc += putRR_SOA(AVStr(RRbuf),(char**)&rrp,name);
 			anc += putRR_PTR(AVStr(RRbuf),(char**)&rrp,name);
 			break;
 	}
+	if( anc == 0 || qtype == TY_SOA || qtype == TY_SRV ){
+		nauth = putRR_SOA(AVStr(RRbuf),(char**)&rrp,name);
+	}
 
 	xrrp = rrp;
+	if( qtype == TY_SOA || qtype == TY_SRV ){
+		unknown = (anc == 0 && nauth == 0);
+	}else{
 	unknown = anc==0 && putRR_A(AVStr(RRbuf),(char**)&xrrp,name)==0 && putRR_PTR(AVStr(RRbuf),(char**)&xrrp,name)==0;
+	}
 	isauth = strtailstr(name,DNS_DOMAIN) != NULL;
 
 	QM = Head.M;
@@ -1399,6 +1652,9 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 		SET_AA(RM,1);
 	else	SET_AA(RM,0);
 
+	if( !RES_proxy() )
+		SET_RD(RM,0);
+	else
 	SET_RD(RM,H_RD(QM));
 	SET_RA(RM,0);
 	if( anc == 0 )
@@ -1411,6 +1667,10 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 		if( unknown ){
 		    if( isauth )
 			SET_RCODE(RM,3); /* UNKNOWN */
+		    else
+		    if( RES_proxy() ){
+			SET_RCODE(RM,3); /* acting as a proxy resolver */
+		    }
 		    else{
 			debug(DBG_FORCE,"%s<-REFUSED %s (DNSCONF=domain:%s)\n",
 				froma,name,DNS_DOMAIN);
@@ -1425,11 +1685,17 @@ int dns_search(PVStr(reply),PCStr(query),int qcc,PCStr(froma),int fromp)
 	putShort((char*)rp,RM);
 	putShort((char*)rp,qdc);
 	putShort((char*)rp,anc);
+/*
 	putShort((char*)rp,0);
+*/
+	putShort((char*)rp,nauth);
 	putShort((char*)rp,0);
 	rp = makeQuestion(QVStr(rp,reply),name,qtype,qclass);
 
+/*
 	if( anc ){
+*/
+	if( anc || nauth ){
 		Bcopy(RRbuf,rp,rrp-RRbuf);
 		rp += (rrp-RRbuf);
 	}
