@@ -181,6 +181,9 @@ static int putUnknownPage(Connection *Conn,FILE *tc,int vno,PCStr(req))
 		return hleng;
 	}
 	inUnknown = getpid();
+	if( strstr(req,"/favicon.ico") ){
+		sv1log("DONT DELAY on unknown /favicon.ico\n");
+	}else
 	delayUnknown(Conn,1,req);
 
 	tmp = TMPFILE("NotFound");
@@ -1191,6 +1194,9 @@ static scanListFunc setbugs(PCStr(value))
 	if( streq(value,"no-keepalive") )
 		HTTP_opts |= HTTP_NOKEEPALIVE;
 	else
+	if( streq(value,"no-keepalive-stls") )
+		HTTP_opts |= HTTP_NOKEEPALIVE_STLS;
+	else
 	if( streq(value,"no-keepalive-proxy") )
 		HTTP_opts |= HTTP_NOKEEPALIVEPROXY;
 	else
