@@ -186,16 +186,8 @@ void putBackTrace(FILE *out)
 
 	Kill(Cpid,9);
 	if( nready == 0 || xpid != cpid ){
-		int rcode;
-		daemonlog("F","\n#### debugger freezed?\n");
 		fprintf(out,"#### error: nready=%d xpid=%d/%d/%d\n",
 			nready,xpid,cpid,Cpid);
-
-		rcode = Kill(cpid,9);
-		sleep(3);
-		xpid = NoHangWait();
-		fprintf(out,"#### Terminate Debugger: kill(%d)=%d xpid=%d\n",
-			cpid,rcode,xpid);
 	}
 }
 
