@@ -21,30 +21,88 @@ History:
 #include <ctype.h>
 #include <errno.h>
 #include <stdio.h>
+#include "ystring.h"
 #include "vsignal.h"
 #include "delegate.h"
-#include "ystring.h"
-extern char *getClientHostPortUser();
+#include "fpoll.h"
+#include "auth.h"
+#include "proc.h"
 
-static char Xdisplay[256];
-static char Xproxy[256];
-static int Xpid;
+extern int IO_TIMEOUT;
 
-static int THREAD;
-static int THREAD_exiting;
-static char *useTHREAD;
-extern char *INHERENT_thread();
+int makeXproxy(Connection *Conn,PVStr(pxdisplay),PCStr(display),PVStr(pxhost),PCStr(relhost),PCStr(me),int timeo);
+extern int *ccx_global;
 
-static int env_valid;
-static jmp_buf tel_env;
-static void sigPIPE(sig){
+static int ccx_telnet(int*ccx,PVStr(buf),int *ip,int cc);
+static int rewriteTelnet(int direction,PVStr(buf),int cc);
+static void putIAC(FILE *tc,int com,int what);
+
+typedef struct {
+	MStr(	te_Xdisplay,256);
+	MStr(	te_Xproxy,256);
+	int	te_Xpid;
+	int	te_THREAD;
+	int	te_THREAD_exiting;
+  const char   *te_useTHREAD;
+	int	te_env_valid;
+	jmp_buf te_tel_env;
+	int	te_gotSIGURG;
+	int	te_dump_commands;
+	int	te_ClientsWill[128];
+	int	te_ServersWill[128];
+	int	te_ClientsDO[256];
+	int	te_docontrol;
+	MStr(	te__opt,8);
+	int	te_keepalive;
+} TelnetEnv;
+static TelnetEnv *telnetEnv;
+#define Xdisplay	telnetEnv->te_Xdisplay
+/**/
+#define Xproxy		telnetEnv->te_Xproxy
+/**/
+#define Xpid		telnetEnv->te_Xpid
+#define THREAD		telnetEnv->te_THREAD
+#define THREAD_exiting	telnetEnv->te_THREAD_exiting
+#define useTHREAD	telnetEnv->te_useTHREAD
+#define env_valid	telnetEnv->te_env_valid
+#define tel_env		telnetEnv->te_tel_env
+#define gotSIGURG	telnetEnv->te_gotSIGURG
+#define dump_commands	telnetEnv->te_dump_commands
+#define ClientsWill	telnetEnv->te_ClientsWill
+#define ServersWill	telnetEnv->te_ServersWill
+#define ClientsDO	telnetEnv->te_ClientsDO
+#define docontrol	telnetEnv->te_docontrol
+#define _opt		telnetEnv->te__opt
+/**/
+#define keepalive	telnetEnv->te_keepalive
+
+void scan_TELNETCONF(Connection*_,PCStr(conf))
+{	CStr(what,32);
+	CStr(value,64);
+	int ival;
+
+	if( telnetEnv == 0 )
+		telnetEnv = NewStruct(TelnetEnv);
+
+	fieldScan(conf,what,value);
+	ival = atoi(value);
+	if( strcaseeq(what,"keepalive") ){
+		if( ival == 0 )
+			ival = 30;
+		keepalive = ival;
+	}
+}
+
+static int guessedHalfdup = -1;
+
+static void sigPIPE(int sig){
 	signal(SIGPIPE,SIG_IGN);
 	if( env_valid ){
 		env_valid = 0;
 		if( !THREAD ) longjmp(tel_env,SIGPIPE);
 	}
 }
-static void sigTERM(sig){
+static void sigTERM(int sig){
 	signal(SIGTERM,SIG_IGN);
 	if( env_valid ){
 		env_valid = 0;
@@ -87,9 +145,10 @@ static void sigTERM(sig){
 #define O_XDISPLOC	35	/* X Display Location */
 #define O_ENVIRON	36	/* Environment Variables */
 
-static int gotSIGURG;
-static void sigurg(sig)
-{	char *ssig;
+static char NOPstr[2] = { IAC, NOP };
+
+static void sigurg(int sig)
+{	const char *ssig;
 
 	signal(SIGURG,sigurg);
 	gotSIGURG += 1;
@@ -112,16 +171,16 @@ setRecvSIGURG(sock)
 }
 */
 
-extern int IO_TIMEOUT;
-
-static Read(what,sock,dst,buf,fsbsize,timeout)
-	char *what,*buf;
+static int Read(PCStr(what),int sock,int dst,PVStr(buf),int fsbsize,int timeout)
 {	int cnt,rcc,ci,cj;
-	char *sdir;
+	const char *sdir;
 	int start,nready;
 
 	rcc = 0;
 	start = time(0);
+	if( keepalive ){
+		timeout = keepalive;
+	}
 	for( cnt = 0; cnt < 10; cnt++ ){
 		if( THREAD )
 			nready = thread_PollIn(sock,timeout*1000);
@@ -139,7 +198,16 @@ static Read(what,sock,dst,buf,fsbsize,timeout)
 			break;
 		}
 		if( nready == 0 ){
+			if( keepalive ){
+				Verbose("%s generate NOP\n",what);
+				rcc = sizeof(NOPstr);
+				Bcopy(NOPstr,buf,rcc);
+				break;
+			}
+
 			if( timeout <= time(0)-start ){
+				sv1log("Timedout %d sec. (by TIMEOUT=io:%ds)\n",
+					timeout,timeout);
 				rcc = -1;
 				break;
 			}
@@ -151,7 +219,7 @@ static Read(what,sock,dst,buf,fsbsize,timeout)
 			continue;
 		}
 		errno = 0;
-		rcc = read(sock,buf,fsbsize);
+		rcc = read(sock,(char*)buf,fsbsize);
 		if( rcc <= 0 ){
 			sv1log("sock=%d read=%d, errno=%d\n",sock,rcc,errno);
 			break;
@@ -161,8 +229,9 @@ static Read(what,sock,dst,buf,fsbsize,timeout)
 			break;
 		}
 
-		if( !gotSIGURG )
+		if( !gotSIGURG ){
 			break;
+		}
 
 		if( rcc <= 0 ){
 			sv1log("%s read failed (%d) waiting SYNCH (%d)\n",
@@ -186,7 +255,7 @@ static Read(what,sock,dst,buf,fsbsize,timeout)
 
 		rcc -= ci;
 		for( cj = 0; cj < rcc; cj++ )
-			buf[cj] = buf[ci+cj];
+			setVStrElem(buf,cj,buf[ci+cj]); /**/
 		break;
 	}
 	for( ci = 0; ci < rcc; ci++ ){
@@ -198,9 +267,6 @@ static Read(what,sock,dst,buf,fsbsize,timeout)
 
 #define C_S	1
 #define S_C	2
-static int dump_commands;
-static int ClientsWill[128];
-static int ServersWill[128];
 
 /*
  * The following stuff is to control ECHOing (and editing with DEL or
@@ -210,6 +276,7 @@ static int ServersWill[128];
  * in RFC857.  But some clients programs (like MS's one) seem to
  * expect the server to be WILL ECHO by default, making full-duplex
  * (character by character) communication without doing local-echo...
+ * (it's the case of WinNT/2K but not the case in Win95/98/Me/XP...)
  * Doing ECHO negotiation between DeleGate and client is a possible
  * solution, like in older version of DeleGate, but it spoiles
  * the transparent relaying between a server and a client.
@@ -220,9 +287,7 @@ static int ServersWill[128];
  * dialogue (its editing) be different from the first one (may not be
  * harmful), and make inconsistency with the server (can be harmful)...
  */
-static int ClientsDO[256];
-static int guessedHalfdup = -1;
-static isHalfdup(){
+static int isHalfdup(){
 	int doga;
 
 	if( doga = ClientsDO[O_SUPAHEAD] ) 
@@ -231,7 +296,7 @@ static isHalfdup(){
 		return 1;
 	return 0; /* it should be 0 based on RFC858... */ 
 }
-static willEcho(){
+static int willEcho(){
 	int doecho;
 
 	/* explicitly ordered by the client */
@@ -247,10 +312,12 @@ static willEcho(){
 		return 0;
 
 	/* it should be 0 baesd on RFC857... */
+	/* 0 for Win95 series but 1 for WinNT series */
 	return 1;
 }
-static logEcho()
-{	char *echo,*supga;
+static void logEcho()
+{	const char *echo;
+	const char *supga;
 
 	switch( ClientsDO[O_ECHO] ){
 		case DO:  echo = "DoEcho"; break;
@@ -267,15 +334,24 @@ static logEcho()
 		isHalfdup(),guessedHalfdup);
 }
 
-static clearServersWill()
+static void clearServersWill()
 {	int si;
 
 	for( si = 0; si < 128; si++)
 		ServersWill[si] = 0;
 }
+static int numServersWill()
+{	int si,num;
 
-static char *code(ch)
-{	char *mn;
+	num = 0;
+	for( si = 0; si < 128; si++ )
+		if( ServersWill[si] )
+			num++;
+	return num;
+}
+
+static const char *code(int ch)
+{	const char *mn;
 	int co;
 
 	co = 1;
@@ -300,8 +376,7 @@ static char *code(ch)
 	}
 	return mn;
 }
-static char _opt[8];
-static char *option(ch)
+static const char *option(int ch)
 {
 	switch( ch ){
 		case O_ECHO:     return "Echo";
@@ -319,26 +394,24 @@ static char *option(ch)
 	return _opt;
 }
 
-static int docontrol;
-extern void *ccx_global;
-static controlcommand(Conn,toC,fromC)
-	Connection *Conn;
-{	char code[2],msg[128];
+static void controlcommand(Connection *Conn,int toC,int fromC)
+{	CStr(code,2);
+	CStr(msg,128);
 
 	write(toC,"\r\n> ",4);
 	read(fromC,code,1);
 	code[1] = 0;
 	sprintf(msg,"%c\r\n",code[0]);
 	write(toC,msg,strlen(msg));
-	global_setCCX(Conn,code,msg);
+	global_setCCX(Conn,AVStr(code),AVStr(msg));
 	write(toC,msg,strlen(msg));
 }
 
-static scanCommands(direction,buf,cc)
-	char *buf;
+static int scanCommands(int direction,PVStr(buf),int cc)
 {	int i,cch,ch,cont;
-	char vch[2048],*mn;
-	char *sdir;
+	CStr(vch,2048);
+	const char *mn;
+	const char *sdir;
 
 	switch( direction ){
 		case CtoS: sdir = "CS"; break;
@@ -351,8 +424,9 @@ static scanCommands(direction,buf,cc)
 
 	for(i = 0; i < cc; i++){
 
-	    if( direction == CtoS )
-		cc += rewriteTelnet(direction,buf+i,cc-i);
+	    if( direction == CtoS ){
+		cc += rewriteTelnet(direction,QVStr(buf+i,buf),cc-i);
+	    }
 
 	    ch = buf[i] & 0xFF;
 	    vch[0] = 0;
@@ -367,7 +441,8 @@ static scanCommands(direction,buf,cc)
 				if( ch == O_XDISPLOC ){
 					if( cch == WONT && Xdisplay[0] ){
 						Verbose("WONT->WILL %s\n",option(ch));
-						buf[i-1] = cch = WILL;
+						cch = WILL;
+						setVStrElem(buf,i-1,cch);
 						mn = code(cch);
 						sprintf(vch,"%-4s %s",mn,option(ch));
 					}
@@ -384,6 +459,9 @@ static scanCommands(direction,buf,cc)
 				ServersWill[ch] = cch;
 			}
 		}else
+		if( ch == NOP ){
+			sprintf(vch,"NOP");
+		}else
 		if( ch == SB ){
 			strcpy(vch,"SB,");
 			ch = buf[++i];
@@ -394,14 +472,14 @@ static scanCommands(direction,buf,cc)
 				if( ' ' < ch && ch < 0x7F ){
 					if( buf[i-1]<=' ' || 0x7F<=buf[i-1] )
 						strcat(vch,",");
-					sprintf(vch+strlen(vch),"%c",ch);
+					Xsprintf(TVStr(vch),"%c",ch);
 				}else
 				if( mn = code(ch) ){
 					strcat(vch,",");
 					strcat(vch,mn);
 				}else{
 					strcat(vch,",");
-					sprintf(vch+strlen(vch),"%d",ch);
+					Xsprintf(TVStr(vch),"%d",ch);
 				}
 				if( ch == SE )
 					break;
@@ -409,15 +487,18 @@ static scanCommands(direction,buf,cc)
 		}
 	        Verbose("%s[%3d] %02x:%3d %s\n",sdir,i,ch,ch,vch);
 	    }else{
+		if( dump_commands & (direction<<8) )
+			sv1log("%s[%3d] %02x:%3d %c\n",
+			sdir,i,ch,ch,(' '<ch&&ch<0x7F)?ch:' ');
 		if( direction == StoC ){
 			/* code conversion */
-			if( ccx_global ) cc = ccx_telnet(ccx_global,buf,&i,cc);
+			if( ccx_global ) cc = ccx_telnet(ccx_global,AVStr(buf),&i,cc);
 		}else
 		if( direction == CtoS ){
-			static numCC;
 			if( ccx_global ){
 				/* conversion switch command ... */
 				/*
+			static numCC;
 				if( ch == ('C'-0x40) ){
 					numCC++;
 					if( 2 < numCC ){
@@ -436,11 +517,9 @@ static scanCommands(direction,buf,cc)
 	}
 	return cc;
 }
-ccx_telnet(ccx,buf,ip,cc)
-	void *ccx;
-	unsigned char *buf;
-	int *ip;
-{	unsigned char out[4096],tmp[4096];
+static int ccx_telnet(int*ccx,PVStr(buf),int *ip,int cc)
+{	CStr(out,4096);
+	CStr(tmp,4096);
 	int i,j,k,l,tc,oc;
 
 	i = *ip;
@@ -448,53 +527,84 @@ ccx_telnet(ccx,buf,ip,cc)
 		if( (buf[j] & 0xFF) == IAC )
 			break;
 
-	oc = CCXexec(ccx,buf+i,j-i,out,sizeof(out));
+	oc = CCXexec((CCXP)ccx,buf+i,j-i,AVStr(out),sizeof(out));
 	tc = 0;
-	for( k = j; k < cc; k++ ) tmp[tc++] = buf[k];
-	for( k = 0; k < oc; k++ ) buf[i++] = out[k];
+	for( k = j; k < cc; k++ ){
+		if( elnumof(tmp) <= tc ){
+			break;
+		}
+		setVStrElemInc(tmp,tc,buf[k]); /**/
+	}
+	for( k = 0; k < oc; k++ ){
+		assertVStr(buf,buf+i+1);
+		setVStrElemInc(buf,i,out[k]); /**/
+	}
 	l = i;
-	for( k = 0; k < tc; k++ ) buf[l++] = tmp[k];
+	for( k = 0; k < tc; k++ ){
+		assertVStr(buf,buf+l+1);
+		setVStrElemInc(buf,l,tmp[k]); /**/
+	}
 	cc = l;
 
 	*ip = i;
 	return cc;
 }
 
-static getCommand(fp,cbuf)
-	FILE *fp;
-	char *cbuf;
+static int getCommand(FILE *fp,UTag *Tcbuf,int off)
 {	int ci,ch,ach;
+	refQStr(cbuf,Tcbuf->ut_addr); /**/
+	int cx;
+
+	cbuf = (char*)Tcbuf->ut_addr + off;
+	cx = Tcbuf->ut_size - off - 1;
 
 	ci = 0;
-	cbuf[ci++] = ch = getc(fp);
+	ch = getc(fp);
+	setVStrElemInc(cbuf,ci,ch);
 
 	if( WILL <= ch && ch <= DONT ){
-		cbuf[ci++] = ach = getc(fp);
+		ach = getc(fp);
+		setVStrElemInc(cbuf,ci,ach);
 	}else
 	if( ch == SB ){
 		do {
+			if( cx <= ci ){
+				notify_overflow("Telnet Command",Tcbuf->ut_addr,off+ci);
+				discardBuffered(fp);
+				break;
+			}
 			ch = getc(fp);
 			if( ch == EOF )
 				break;
-			cbuf[ci++] = ch;
+			setVStrElemInc(cbuf,ci,ch);
 		} while(ch != SE);
 	}
 	return ci;
 }
-static getline(visible,fc,tc,line,cbuf)
-	FILE *fc,*tc;
-	char *line,*cbuf;
+/**/
+static int xgetline(void *visible,FILE *fc,FILE *tc,UTag *Tline,UTag *Tcbuf)
 {	int ready;
 	int ch,ci;
 	int endCR;
 	int lx;
 	int ii;
-	char *cb;
+	const char *cb;
 	int timeout;
-	char cbufbuf[1024];
+	CStr(cbufbuf,1024);
+	refQStr(line,Tline->ut_addr); /**/
+	defQStr(cbuf); /*indirect*//**/
+	int lsiz,csiz;
+	UTag Tcbufb;
 
-	if( cbuf == NULL )
-		cbuf = cbufbuf;
+	line = (char*)Tline->ut_addr;
+	lsiz = Tline->ut_size;
+	if( Tcbuf == NULL ){
+		Tcbuf = &Tcbufb;
+		setQStr(Tcbuf->ut_addr,cbufbuf,sizeof(cbufbuf));
+		Tcbuf->ut_size = sizeof(cbufbuf);
+	}
+	setQStr(cbuf,Tcbuf->ut_addr,Tcbuf->ut_size);
+	csiz = Tcbuf->ut_size;
 	endCR = 0;
 	ci = 0;
 	lx = 0;
@@ -516,13 +626,23 @@ static getline(visible,fc,tc,line,cbuf)
 	}
 
 	for(;;){
+		if( lsiz-1 <= lx ){
+			notify_overflow("Telnet Input",line,lx);
+			discardBuffered(fc);
+			break;
+		}
+		if( csiz-1 <= ci ){
+			notify_overflow("Telnet Command",cbuf,ci);
+			discardBuffered(fc);
+			break;
+		}
 		if( fflush(tc) == EOF )
 			break;
 
 		ready = fPollIn(fc,timeout);
 		if( ready <= 0 ){
 			if( ready == 0 ){
-fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
+ fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
 					timeout);
 				fflush(tc);
 				return EOF;
@@ -550,7 +670,7 @@ fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
 
 			if( ch == '?' ){
 				if( fPollIn(fc,300) == 0 ){
-					line[lx++] = ch;
+					setVStrElemInc(line,lx,ch);
 					goto EOL;
 				}
 			}
@@ -566,9 +686,10 @@ fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
 
 		switch( ch ){
 			case IAC:
-				cbuf[ci] = ch;
-				ii = 1 + getCommand(fc,cbuf+ci+1);
-				ci += scanCommands(CtoS,cbuf+ci,ii);
+				setVStrElem(cbuf,ci,ch);
+				ii = 1 + getCommand(fc,Tcbuf,ci+1);
+				ci += scanCommands(CtoS,
+					QVStr(cbuf+ci,Tcbuf->ut_addr),ii);
 				break;
 
 			case 0: break;
@@ -583,6 +704,7 @@ fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
 			case 0x7F:
 			case 'H'-0x40:
 				if( 0 < lx ){
+					if( willEcho() )
 					fwrite("\b \b",1,3,tc);
 					lx--;
 				}
@@ -610,59 +732,56 @@ fprintf(tc,"\r\n---------- PROXY-TELNET login: TIMEOUT(%d).\r\n",
 				Verbose("EOL char = %x\n",ch);
 				goto EOL;
 			default:
-				line[lx++] = ch;
+				setVStrElemInc(line,lx,ch);
 				break;
 		}
 	} EOL:
-	line[lx] = 0;
+	setVStrEnd(line,lx);
 
 	if( willEcho() )
 	fputs("\r\n",tc);
+
+	Tcbuf->ut_leng = ci;
 	return ci;
 }
 
-static putIAC(tc,com,what)
-	FILE *tc;
-{	char ibuf[32],ii;
+static void putIAC(FILE *tc,int com,int what)
+{	CStr(ibuf,32);
+	char ii;
 
 	ii = 0;
-	ibuf[ii++] = IAC; ibuf[ii++] = com; ibuf[ii++] = what;
-	ii = scanCommands(DtoC,ibuf,ii);
+	ibuf[ii++] = IAC; ibuf[ii++] = com; ibuf[ii++] = what; /**/
+	ii = scanCommands(DtoC,AVStr(ibuf),ii);
 	fwrite(ibuf,1,ii,tc);
 }
-static sayWelcome(Conn,tc)
-	Connection *Conn;
-	FILE *tc;
-{	char *aurl,rurl[256],msg[2048];
+static void sayWelcome(Connection *Conn,FILE *tc)
+{	const char *aurl;
+	CStr(rurl,256);
+	CStr(msg,2048);
 
 	aurl = "/-/builtin/mssgs/telnet/telnet-banner.dhtml";
-	getBuiltinData(Conn,"TELNET-banner",aurl,msg,sizeof(msg),rurl);
+	getBuiltinData(Conn,"TELNET-banner",aurl,AVStr(msg),sizeof(msg),AVStr(rurl));
 	put_eval_dhtml(Conn,rurl,tc,msg);
 }
-static putHELP(Conn,tc)
-	Connection *Conn;
-	FILE *tc;
-{	char *aurl,rurl[256],msg[2048];
+static void putHELP(Connection *Conn,FILE *tc)
+{	const char *aurl;
+	CStr(rurl,256);
+	CStr(msg,2048);
 
 	aurl = "/-/builtin/mssgs/telnet/telnet-help.dhtml";
-	getBuiltinData(Conn,"TELNET-help",aurl,msg,sizeof(msg),rurl);
+	getBuiltinData(Conn,"TELNET-help",aurl,AVStr(msg),sizeof(msg),AVStr(rurl));
 	put_eval_dhtml(Conn,rurl,tc,msg);
 }
-static cantConnMessage(Conn,serv,tc)
-	Connection *Conn;
-	char *serv;
-	FILE *tc;
-{	char *aurl,rurl[256],msg[2048];
+static void cantConnMessage(Connection *Conn,PCStr(serv),FILE *tc)
+{	const char *aurl;
+	CStr(rurl,256);
+	CStr(msg,2048);
 
 	aurl = "/-/builtin/mssgs/telnet/telnet-cantconn.dhtml";
-	getBuiltinData(Conn,"TELNET-cantconn",aurl,msg,sizeof(msg),rurl);
+	getBuiltinData(Conn,"TELNET-cantconn",aurl,AVStr(msg),sizeof(msg),AVStr(rurl));
 	put_eval_dhtml(Conn,rurl,tc,msg);
 }
-static ConnectedMessage(Conn,serv,tc,comline)
-	Connection *Conn;
-	char *serv;
-	FILE *tc;
-	char *comline;
+static void ConnectedMessage(Connection *Conn,PCStr(serv),FILE *tc,PCStr(comline))
 {
 	fprintf(tc,"-- Connected to %s.\r\n",serv);
 	if( toProxy ){
@@ -672,13 +791,9 @@ static ConnectedMessage(Conn,serv,tc,comline)
 	fflush(tc);
 }
 
-static get_hostname(Conn,fc,tc,line,cbuf,csize,ncom)
-	Connection *Conn;
-	FILE *fc,*tc;
-	char *line,*cbuf;
-	int *csize;
+static int get_hostname(Connection *Conn,FILE *fc,FILE *tc,UTag *Tline,UTag *Tcbuf,int ncom)
 {	int ci;
-	char clhost[256];
+	CStr(clhost,256);
 
 /*
 	putIAC(tc,WILL,O_ECHO);
@@ -689,8 +804,8 @@ static get_hostname(Conn,fc,tc,line,cbuf,csize,ncom)
 		sayWelcome(Conn,tc);
 
 	if( ncom == 0 )
-	if( !source_permitted(Conn) ){
-		getpeerNAME(FromC,clhost);
+	if( !source_permittedX(Conn) ){
+		getpeerNAME(FromC,AVStr(clhost));
 		fprintf(tc,
 		"!!!!!!!! HOST <%s> not permitted by DeleGate\r\n",clhost);
 		fflush(tc);
@@ -701,16 +816,14 @@ static get_hostname(Conn,fc,tc,line,cbuf,csize,ncom)
 	fprintf(tc,">> Host name: ");
 	if( fflush(tc) == EOF )
 		return EOF;
-	ci = getline(1,fc,tc,line,cbuf);
+	ci = xgetline((void*)1,fc,tc,Tline,Tcbuf);
 	if( ci == EOF )
 		return EOF;
 
 	fflush(tc);
-	*csize = ci;
 	return 0;
 }
-static relayClientsWill(ts,op)
-	FILE *ts;
+static void relayClientsWill(FILE *ts,int op)
 {	int co;
 
 	if( co = ClientsWill[op] ){
@@ -720,16 +833,14 @@ static relayClientsWill(ts,op)
 		putc(op,ts);
 	}
 }
-static relayClientsWills(ts)
-	FILE *ts;
+static void relayClientsWills(FILE *ts)
 {	int op;
 	for( op = 0; op < 128; op++ )
 		relayClientsWill(ts,op);
 }
 
-static CS(Conn,fcbsize,omask)
-	Connection *Conn;
-{	char buf[0x4000];
+static int CS(Connection *Conn,int fcbsize,int omask)
+{	CStr(buf,0x4000);
 	int gotsig;
 	int rcc;
 	int count,total;
@@ -749,12 +860,13 @@ static CS(Conn,fcbsize,omask)
 		env_valid = 1;
 		sigsetmask(omask);
 		for(;;){
-			rcc = Read("CS",FromC,ToS,buf,fcbsize,IO_TIMEOUT);
+			rcc = Read("CS",FromC,ToS,AVStr(buf),fcbsize,IO_TIMEOUT);
 			if( rcc <= 0 )
 				break;
 
-			if( dump_commands && C_S )
-				rcc = scanCommands(CtoS,buf,rcc);
+			if( dump_commands && C_S ){
+				rcc = scanCommands(CtoS,AVStr(buf),rcc);
+			}
 
 			count += 1;
 			total += rcc;
@@ -774,6 +886,7 @@ static CS(Conn,fcbsize,omask)
 		}
 		if( rcc <= 0 ) sv1log("CS-EOF\n");
 	}
+	env_valid = 0;
 	sv1log("CS-RELAY[%d>%d]: %dBytes %dI/O buf=%d\n",
 		FromC,ToS,total,count,fcbsize);
 
@@ -786,9 +899,8 @@ static CS(Conn,fcbsize,omask)
 	}
 	return total;
 }
-static SC(Conn,fsbsize,omask)
-	Connection *Conn;
-{	char buf[0x4000];
+static int SC(Connection *Conn,int fsbsize,int omask)
+{	CStr(buf,0x4000);
 	int gotsig;
 	int rcc,wcc;
 	int count,total;
@@ -803,27 +915,32 @@ static SC(Conn,fsbsize,omask)
 		env_valid = 1;
 		sigsetmask(omask);
 		for(;;){
-			rcc = Read("SC",FromS,ToC,buf,fsbsize,IO_TIMEOUT);
+			rcc = Read("SC",FromS,ToC,AVStr(buf),fsbsize,IO_TIMEOUT);
 			if( rcc <= 0 )
 				break;
 
 			count += 1;
 			total += rcc;
-			if( (dump_commands & S_C) || CCX )
-				rcc = scanCommands(StoC,buf,rcc);
+			if( (dump_commands & S_C) || CCX ){
+				rcc = scanCommands(StoC,AVStr(buf),rcc);
+				if( rcc == 0 ){
+					/* ESC seq. pending in CCX */
+					continue;
+				}
+			}
 
 			if( (wcc = write(ToC,buf,rcc)) <= 0 )
 				break;
 		}
 		if( rcc <= 0 ) sv1log("SC-EOF\n");
 	}
+	env_valid = 0;
 	sv1log("SC-RELAY[%d<%d]: %dBytes %dI/O buf=%d\n",
 		ToC,FromS,total,count,fsbsize);
 	return total;
 }
 
-static THREADexit(Conn)
-	Connection *Conn;
+static void THREADexit(Connection *Conn)
 {	int efd;
 
 	if( THREAD && THREAD_exiting == 0 ){
@@ -841,10 +958,21 @@ static THREADexit(Conn)
 		if( efd != ToC   ) dup2(efd,ToC);
 		if( efd != FromS ) dup2(efd,FromS);
 		if( efd != ToS   ) dup2(efd,ToS);
+
+		/*
+		 * wait the child thread to exit on closed socket {From,To}{C,S}
+		 * no return if this is the child thread.
+		 */
+		msleep(10);
+		if( THREAD_exiting != 2 )
+			sv1log("THREAD_exit: failed, exiting=%d\n",THREAD_exiting);
+	}
+	else
+	if( THREAD ){
+		THREAD_exiting++;
 	}
 }
-static SC1(Conn,fsbsize,omask)
-	Connection *Conn;
+static int SC1(Connection *Conn,int fsbsize,int omask)
 {	int cc;
 
 	dumpTimer();
@@ -855,8 +983,7 @@ static SC1(Conn,fsbsize,omask)
 	THREADexit(Conn);
 	return cc;
 }
-static CS1(Conn,fcbsize,omask)
-	Connection *Conn;
+static int CS1(Connection *Conn,int fcbsize,int omask)
 {	int cc;
 
 	dumpTimer();
@@ -868,14 +995,22 @@ static CS1(Conn,fcbsize,omask)
 	return cc;
 }
 
-static bidirectional_relay(Conn,fcbsize,fsbsize)
-	Connection *Conn;
+static int bidirectional_relay(Connection *Conn,int fcbsize,int fsbsize)
 {	register int ppid,cpid;
 	int omask;
 	int total;
-	void (*opipe)();
-	void (*oterm)();
-	void (*ointr)();
+	void (*opipe)(int);
+	void (*oterm)(int);
+	void (*ointr)(int);
+
+	CStr(snoop,64);
+	int sf;
+	if( 0 <= find_CMAP(Conn,"DUMP",AVStr(snoop)) ){
+		sf = 0;
+		if( strcasestr(snoop,"tosv") ) sf |= C_S|(CtoS<<8);
+		if( strcasestr(snoop,"tocl") ) sf |= S_C|(StoC<<8);
+		dump_commands |= sf;
+	}
 
 	useTHREAD = INHERENT_thread();
 
@@ -897,8 +1032,8 @@ static bidirectional_relay(Conn,fcbsize,fsbsize)
 	if( useTHREAD ){
 		THREAD = 1;
 		THREAD_exiting = 0;
-		thread_fork(0,CS1,Conn,fsbsize,omask);
-		SC1(Conn,fcbsize,omask,ppid);
+		thread_fork(0,(IFUNCP)CS1,Conn,fsbsize,omask);
+		SC1(Conn,fcbsize,omask /*,ppid*/);
 		THREAD = 0;
 	}else{
 		ppid = getpid();
@@ -924,24 +1059,22 @@ static bidirectional_relay(Conn,fcbsize,fsbsize)
 	return total;
 }
 
-static getServersWill(fromS,tc)
-	FILE *tc;
+static void getServersWill(int fromS,FILE *tc)
 {	int rcc;
-	char scbuf[4096];
+	CStr(scbuf,4096);
 
 	if( PollIn(fromS,1000) ){
-		rcc = readTIMEOUT(fromS,scbuf,sizeof(scbuf));
+		rcc = readTIMEOUT(fromS,AVStr(scbuf),sizeof(scbuf));
 		sv1log("######## %d bytes from the server\n",rcc);
-		rcc = scanCommands(StoC,scbuf,rcc);
+		rcc = scanCommands(StoC,AVStr(scbuf),rcc);
 		fwrite(scbuf,1,rcc,tc);
 		fflush(tc);
 	}
 }
 
-static putXdisplay(ts,display)
-	FILE *ts;
-	char *display;
-{	char ibuf[128],ch;
+static int putXdisplay(FILE *ts,PCStr(display))
+{	CStr(ibuf,128);
+	char ch;
 	int ii,ij;
 	int opt;
 
@@ -956,34 +1089,37 @@ if( opt == O_ENVIRON )
 	return 0;
 
 	ii = 0;
-	ibuf[ii++] = IAC; ibuf[ii++] = WILL; ibuf[ii++] = opt;
-	ibuf[ii++] = IAC; ibuf[ii++] = SB;   ibuf[ii++] = opt;
+	ibuf[ii++] = IAC; ibuf[ii++] = WILL; ibuf[ii++] = opt; /**/
+	ibuf[ii++] = IAC; ibuf[ii++] = SB;   ibuf[ii++] = opt; /**/
 	ibuf[ii++] = IS;
 	if( opt == O_ENVIRON ){
 		ibuf[ii++] = VAR;
-		for( ij = 0; ch = "DISPLAY"[ij]; ij++ )
-			ibuf[ii++] = ch;
+		for( ij = 0; ch = "DISPLAY"[ij]; ij++ ){
+			ibuf[ii++] = ch; /**/
+		}
 		ibuf[ii++] = VALUE;
 	}
-	for( ij = 0; ch = display[ij]; ij++ )
-		ibuf[ii++] = ch;
+	for( ij = 0; ch = display[ij]; ij++ ){
+		if( elnumof(ibuf) <= ii+3 )
+			break;
+		ibuf[ii++] = ch; /**/
+	}
 	ibuf[ii++] = IAC; ibuf[ii++] = SE;
-	ii = scanCommands(DtoC,ibuf,ii);
+	ii = scanCommands(DtoC,AVStr(ibuf),ii);
 	fwrite(ibuf,1,ii,ts);
 	return 1;
 }
 
-static makeXproxy1(Conn,myname,peername)
-	Connection *Conn;
-	char *myname,*peername;
-{	char me[256],myhp[256];
+static int makeXproxy1(Connection *Conn,PVStr(myname),PCStr(peername),int peerport)
+{	CStr(me,256);
+	CStr(myhp,256);
 	int xpid;
 
 /*
 Socks can accept only one X connection ...
 
 	if( Conn->sv_viaSocks ){
-		char rhost[32];
+		CStr(rhost,32);
 		int rport,rsock;
 
 		rsock = bindViaSocks(Conn,DST_HOST,DST_PORT,rhost,&rport);
@@ -992,60 +1128,46 @@ Socks can accept only one X connection ...
 	}
 */
 
-	ClientIF_HP(Conn,myhp);
+	ClientIF_HP(Conn,AVStr(myhp));
 	sprintf(me,"%s://%s/",DFLT_PROTO,myhp);
-	xpid = makeXproxy(Conn,Xproxy,Xdisplay,myname,peername,me,0);
+	xpid = makeXproxy(Conn,AVStr(Xproxy),Xdisplay,AVStr(myname),peername,me,0);
+
+	/* if SRCIF="proxyDisplay:*:X" is given */{
+		CStr(xdisp,256);
+		const char *dp;
+		int xport;
+
+		if( SRCIFfor(Conn,"X",peername,peerport,AVStr(xdisp),&xport) ){
+			if( dp = strchr(Xproxy,':') ){
+				sv1log("#### Xproxy = %s (%s)\n",xdisp,Xproxy);
+				Strrplc(AVStr(Xproxy),dp-Xproxy,xdisp);
+			}
+		}
+	}
+
 	sv1log("#### Xproxy[%d]: %s <- %s <- %s\n",
 		xpid,Xdisplay,Xproxy,peername);
 	return xpid;
 }
 
-extern char *gethostaddr();
-static char *xhosts[64];
-static setXhost(host,hosts)
-	char *host,*hosts;
-{	int xi,xj,rm;
-	char *host1;
-	
-	rm = 0;
-	if( *host == '+' )
-		host++;
-	else
-	if( *host == '-' ){
-		rm = 1;
-		host++;
-	}
-	for( xi = 0; host1 = xhosts[xi]; xi++ ){
-		if( streq(host1,host) )
-			break;
-	}
-	if( rm ){
-		if( host1 != NULL ){ /* found */
-			free(host1);
-			for( xj = xi; xhosts[xj] = xhosts[xj+1]; xi++){}
-		}
-	}else{
-		if( host1 == NULL ){
-			xhosts[xi] = strdup(host);
-		}
-	}
-	hosts[0] = 0;
-	for( xi = 0; host1 = xhosts[xi]; xi++ ){
-		if( 0 < xi )
-			strcat(hosts," ");
-		strcat(hosts,host1);
-	}
-}
-
-static proxy_telnet(Conn)
-	Connection *Conn;
+static void proxy_telnet(Connection *Conn)
 {	FILE *tc,*fc,*ts;
-	char comline[256],cbuf[256];
-	char command[256],hostport[256],serv[256],*addr,clnt[256];
-	char auth[256],auser[256],ahost[256],*iuser;
+	CStr(comline,256);
+	CStr(cbuf,256);
+	CStr(command,256);
+	CStr(hostport,256);
+	CStr(serv,256);
+	CStr(clnt,256);
+	const char *addr;
+	CStr(auth,256);
+	CStr(auser,256);
+	CStr(ahost,256);
+	const char *iuser;
 	int port;
 	int csize;
 	int ns,ncom;
+	const char *mount_opts;
+	UTag Tcomline,Tcbuf;
 
 	tc = fdopen(ToC,"w");
 	fc = fdopen(ToC,"r");
@@ -1054,12 +1176,18 @@ static proxy_telnet(Conn)
 	Xdisplay[0] = 0;
 	dump_commands |= C_S; /* to check WILL/WONT TERMTYPE ... */
 
+	setQStr(Tcomline.ut_addr,comline,sizeof(comline));
+	Tcomline.ut_size = sizeof(comline);
+	setQStr(Tcbuf.ut_addr,cbuf,sizeof(cbuf));
+	Tcbuf.ut_size = sizeof(cbuf);
+
 	ncom = 0;
 	for(ns = 0;;ns++){
-		if( get_hostname(Conn,fc,tc,comline,cbuf,&csize,ncom) == EOF ){
+		if( get_hostname(Conn,fc,tc,&Tcomline,&Tcbuf,ncom) == EOF ){
 			sv1log("EOF from the client\n");
 			break;
 		}
+		csize = Tcbuf.ut_leng;
 		ncom++;
 
 		if( comline[0] == 0 ){
@@ -1067,8 +1195,17 @@ static proxy_telnet(Conn)
 			break;
 		}
 
+		if( mount_opts = CTX_mount_url_to(Conn,NULL,"GET",AVStr(comline)) ){
+			if( strncasecmp(comline,"telnet://",9) == 0 ){
+				const char *dp;
+				wordscanY(comline+9,AVStr(comline),sizeof(comline),"^/? \t\r\n");
+				if( dp = strchr(comline,':') )
+					*(char*)dp = ' '; /**/
+			}
+		}
+
 		command[0] = hostport[0] = 0;
-		sscanf(comline,"%s %[^\r\n]",command,hostport);
+		Xsscanf(comline,"%s %[^\r\n]",AVStr(command),AVStr(hostport));
 		if( streq(command,"q") || streq(command,"quit") || streq(command,"exit") ){
 			sv1log("[%s] command from the client\n",command);
 			break;
@@ -1079,14 +1216,14 @@ static proxy_telnet(Conn)
 			continue;
 		}else
 		if( streq(command,"x") || streq(command,"x-gw") ){
-			char host[128];
+			CStr(host,128);
 			int port;
 
 			sv1log("TIS compati. proxy-telnet with X-proxy\n");
 			host[0] = 0;
-			if( sscanf(hostport,"%[^:]:%d",host,&port) != 2 ){
+			if( Xsscanf(hostport,"%[^:]:%d",AVStr(host),&port) != 2 ){
 				if( host[0] == 0 )
-					getpeerNAME(FromC,hostport);
+					getpeerNAME(FromC,AVStr(hostport));
 				strcat(hostport,":0");
 				fprintf(tc,"####[regarded as] x-gw %s\r\n",hostport);
 				fflush(tc);
@@ -1095,15 +1232,14 @@ static proxy_telnet(Conn)
 			continue;
 		}else
 		if( streq(command,"a") || streq(command,"accept") ){
-			char myname[256];
-			char hosts[1024];
+			CStr(myname,256);
 
 			if( hostport[0] == 0 ){
 				fprintf(tc,"???? Usage: accept hostname\r\n");
 				fflush(tc);
 				continue;
 			}
-			hostIFfor(hostport,myname);
+			hostIFfor(hostport,AVStr(myname));
 			sv1log("myname = %s\n",myname);
 			if( myname[0] == 0 ){
 				fprintf(tc,"???? cannot find a route to %s\r\n",hostport);
@@ -1111,19 +1247,18 @@ static proxy_telnet(Conn)
 				continue;
 			}
 
-			/*setXhost(hostport,hosts);*/
 			if( 0 < Xpid )
 				Kill(Xpid,SIGTERM);
 
-			Xpid = makeXproxy1(Conn,myname,hostport);
+			Xpid = makeXproxy1(Conn,AVStr(myname),hostport,0);
 			fprintf(tc,"####[X-Proxy] accept host '%s'\r\n",hostport);
 			fprintf(tc,"####[started] use DISPLAY '%s'\r\n",Xproxy);
 			fflush(tc);
 			continue;
 		}else
 		if( streq(command,"j") ){
-			char stat[256];
-			global_setCCX(Conn,hostport,stat);
+			CStr(stat,256);
+			global_setCCX(Conn,AVStr(hostport),AVStr(stat));
 			fprintf(tc,"-- charcode conversion [%s]%s\r\n",
 				hostport,stat);
 			fflush(tc);
@@ -1137,7 +1272,7 @@ static proxy_telnet(Conn)
 		}
 		serv[0] = 0;
 		port = DFLT_PORT;
-		sscanf(hostport,"%s %d",serv,&port);
+		Xsscanf(hostport,"%s %d",AVStr(serv),&port);
 
 		if( serv[0] == 0 ){
 			sv1log("empty command for QUIT form the client\n");
@@ -1145,11 +1280,17 @@ static proxy_telnet(Conn)
 		}
 
 		auser[0] = ahost[0] = 0;
-		if( doAUTH(Conn,fc,tc,"telnet",serv,port,auser,ahost,getline,NULL) == EOF )
+		if( doAUTH0(Conn,fc,tc,"telnet",serv,port,AVStr(auser),AVStr(ahost),(iFUNCP)xgetline,NULL) == EOF )
 			break;
+		/* DFLT_HOST and DFLT_PORT is set as side effect in doAUTH0() */
 
-		if( iuser = getClientHostPortUser(Conn,clnt,NULL) )
-			sprintf(clnt+strlen(clnt),"(%s)",iuser);
+		if( Xdisplay[0] ){
+			/* suppress direct connection by MASTER */
+			strcpy(D_SERVER,"telnet://-/");
+		}
+
+		if( iuser = getClientHostPortUser(Conn,AVStr(clnt),NULL) )
+			Xsprintf(TVStr(clnt),"(%s)",iuser);
 		auth[0] = 0;
 		if( auser[0] )
 			sprintf(auth,"<%s@%s>",auser,ahost);
@@ -1160,6 +1301,7 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 		if( (addr = gethostaddr(serv)) == NULL )
 			addr = "unknown host";
 
+		if( mount_opts == NULL )
 		fprintf(tc,"-- Trying %s [%s:%d] ...\r\n",
 			serv,addr,DST_PORT);
 		fflush(tc);
@@ -1180,13 +1322,24 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 				fflush(ts);
 			}
 			if( Xdisplay[0] ){
-				char myname[256],peername[256];
+				CStr(myname,256);
+				CStr(peername,256);
+				int port;
 
-				gethostNAME(ToS,myname);
+				gethostNAME(ToS,AVStr(myname));
+				/*
 				getpeerNAME(ToS,peername);
 				Xpid = makeXproxy1(Conn,myname,peername);
+				*/
+				port = getpeerNAME(ToS,AVStr(peername));
+				Xpid = makeXproxy1(Conn,AVStr(myname),peername,port);
 			}
+/*
 			if( toProxy ){
+*/
+			if( toProxy || toMaster && Xdisplay[0] ){
+				putIAC(ts,DO,O_ECHO); /* enable server ECHO */
+
 				sv1log("#### connected to Proxy telnet\n");
 				/* fprintf(ts,"telnet %s\r\n",serv); */
 				if( Xdisplay[0] ){
@@ -1194,6 +1347,8 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 					fflush(ts);
 				}
 				fprintf(ts,"%s",comline);
+				if( toMaster )
+					fprintf(ts,"\r\n");
 			}else{
 				if( Xdisplay[0] ){
 					int set;
@@ -1208,7 +1363,19 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 			/* Some telnet client (at least SunOS's one)
 			 * does not repeat WILL/WONT twice.
 			 */
+			if( numServersWill() != 0 ){ /* is Telnet server */
 			relayClientsWills(ts);
+			}else{
+			/* how the client can be reset to initial state ? */
+				/*
+				sv1log("#### force kludge line mode ...\n");
+				putIAC(tc,WONT,O_ECHO);
+				putIAC(tc,WILL,O_SUPAHEAD);
+				putIAC(tc,DO,O_TIMING);
+				... skip until TIMING-MARK ...
+				*/
+			}
+
 			fclose(ts);
 			bidirectional_relay(Conn,1024,1024);
 			fprintf(tc,"\r\n");
@@ -1218,6 +1385,9 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 				Xpid = 0;
 			}
 			ncom = 0;
+			if( ImMaster ){
+				break;
+			}
 		}
 	}
 	if( 0 < Xpid ){
@@ -1226,17 +1396,22 @@ fputLog(Conn,"Login","TELNET; from=%s%s; to=%s\n",clnt,auth,serv);
 	}
 }
 
-static AsServer(Conn)
-	Connection *Conn;
+static void AsServer(Connection *Conn)
 {	FILE *fc,*tc,*fp;
-	char uname[128],myhost[128];
-	char user[128],pass[128];
-	char ibuf[128];
+	CStr(uname,128);
+	CStr(myhost,128);
+	CStr(user,128);
+	CStr(pass,128);
+	CStr(ibuf,128);
 	int ii;
-	char line[1024],*dp,com[128],*arg,pwd[1024];
+	CStr(line,1024);
+	const char *dp;
+	CStr(com,128);
+	const char *arg;
+	CStr(pwd,1024);
 
-	ClientIF_name(Conn,fileno(fc),myhost);
-	Uname(uname);
+	ClientIF_name(Conn,FromC,AVStr(myhost));
+	Uname(AVStr(uname));
 
 	fc = fdopen(FromC,"r");
 	tc = fdopen(ToC,"w");
@@ -1249,9 +1424,9 @@ static AsServer(Conn)
 	ibuf[ii++] = IAC; ibuf[ii++] = DO;   ibuf[ii++] = O_ECHO;
 	fwrite(ibuf,1,ii,tc);
 	fflush(tc);
-	scanCommands(DtoC,ibuf,ii);
+	scanCommands(DtoC,AVStr(ibuf),ii);
 	ii = read(fileno(fc),line,sizeof(line));
-	scanCommands(CtoD,ibuf,ii);
+	scanCommands(CtoD,AVStr(ibuf),ii);
 
 	fputs("UserName: ",tc);
 	fflush(tc);
@@ -1265,9 +1440,9 @@ static AsServer(Conn)
 	ibuf[ii++] = IAC; ibuf[ii++] = DONT; ibuf[ii++] = O_ECHO;
 	fwrite(ibuf,1,ii,tc);
 	fflush(tc);
-	scanCommands(DtoC,ibuf,ii);
+	scanCommands(DtoC,AVStr(ibuf),ii);
 	ii = read(fileno(fc),line,sizeof(line));
-	scanCommands(CtoD,ibuf,ii);
+	scanCommands(CtoD,AVStr(ibuf),ii);
 
 	fputs("PassWord: ",tc);
 	fflush(tc);
@@ -1283,9 +1458,9 @@ static AsServer(Conn)
 	ibuf[ii++] = IAC; ibuf[ii++] = DO;   ibuf[ii++] = O_ECHO;
 	fwrite(ibuf,1,ii,tc);
 	fflush(tc);
-	scanCommands(DtoC,ibuf,ii);
+	scanCommands(DtoC,AVStr(ibuf),ii);
 	ii = read(fileno(fc),line,sizeof(line));
-	scanCommands(CtoD,ibuf,ii);
+	scanCommands(CtoD,AVStr(ibuf),ii);
 
 	for(;;){
 		getcwd(pwd,sizeof(pwd));
@@ -1296,7 +1471,7 @@ static AsServer(Conn)
 			break;
 		sv1log("%s> %s",pwd,line);
 		if( dp = strpbrk(line,"\r\n") )
-			*dp = 0;
+			truncVStr(dp);
 		arg = wordScan(line,com);
 		while( *arg == ' ' || *arg == '\t' )
 			arg++;
@@ -1310,7 +1485,7 @@ static AsServer(Conn)
 		fp = popen(line,"r");
 		while( fgets(line,sizeof(line),fp) != NULL ){
 			if( dp = strpbrk(line,"\r\n") )
-				*dp = 0;
+				truncVStr(dp);
 			fputs(line,tc);
 			fputs("\r\n",tc);
 		}
@@ -1323,9 +1498,26 @@ EXIT:
 	fclose(fc);
 }
 
-service_telnet(Conn)
-	Connection *Conn;
+/*
+ * relaying to non-Telnet arbitrary port is enabled with REMITTABLE="telnet"
+ */
+static int telnetonly(Connection *Conn)
+{	int relayany;
+	Port sv;
+
+	sv = Conn->sv;
+	strcpy(REAL_HOST,"-");
+	strcpy(REAL_PROTO,"telnet");
+	REAL_PORT = 99999;
+	relayany = service_permitted2(Conn,"telnet",1);
+	Conn->sv = sv;
+	return !relayany;
+}
+int service_telnet(Connection *Conn)
 {
+	if( telnetEnv == 0 )
+		telnetEnv = NewStruct(TelnetEnv);
+
 	if( LOG_VERBOSE )
 		dump_commands = C_S | S_C;
 
@@ -1333,15 +1525,49 @@ service_telnet(Conn)
 	if( strcmp(iSERVER_PROTO,"telnet") == 0 )
 	if( strcmp(DFLT_HOST,"-.-") == 0 ){
 		AsServer(Conn);
-		return;
+		return 0;
 	}
 	if( isMYSELF(DFLT_HOST) ){
 		ImProxy = 1;
-		return proxy_telnet(Conn);
+		if( PollIn(FromC,10) == 0 ){
+			/* disable local echo in telnet client on Win95 series,
+			 * but don't disable local echo necessary for
+			 * non-Telnet protocols.
+			 */
+			if( telnetonly(Conn) ){
+				write(ToC,NOPstr,2);
+				scanCommands(DtoC,CVStr(NOPstr),2);
+			}
+		}
+		proxy_telnet(Conn);
+		return 0;
 	}
 
+	if( CTX_auth(Conn,NULL,NULL) ) /* with AUTHORIZER */
+	{
+		FILE *fc = fdopen(FromC,"r");
+		FILE *tc = fdopen(ToC,"w");
+		CStr(auser,256);
+		CStr(ahost,256);
+		int dport = DFLT_PORT;
+
+		write(ToC,NOPstr,2);
+		scanCommands(DtoC,CVStr(NOPstr),2);
+
+		auser[0] = ahost[0] = 0;
+		DFLT_PORT = 0; /* to escape "already authorized" */
+		if( doAUTH0(Conn,fc,tc,"telnet",DST_HOST,dport,AVStr(auser),AVStr(ahost),(iFUNCP)xgetline,NULL) == EOF ){
+			return -1;
+		}
+		DFLT_PORT = dport;
+		fcloseFILE(fc);
+		fcloseFILE(tc);
+	}
 	if( ToC < 0 || ToS < 0 )
-		return;
+		connect_to_serv(Conn,FromC,ToC,0);
+
+	if( ToC < 0 || ToS < 0 )
+		return -1;
 
 	/* Telnet clients may not start any negotiation when the
 	 * target port is not the standard telnet port, because the server
@@ -1350,12 +1576,19 @@ service_telnet(Conn)
 	 * negotiation in the relay will be helpful for clients to notice
 	 * that he is connected with telnet server.
 	 */
-	{	char buf[3];
+	if( PollIn(FromS,10) == 0 ){
+		write(ToC,NOPstr,2);
+		scanCommands(DtoC,CVStr(NOPstr),2);
+	}
+	/*
+	{	CStr(buf,3);
 		buf[0] = IAC; buf[1] = WILL; buf[2] = O_SUPAHEAD;
 		write(ToC,buf,3);
-		scanCommands(DtoC,buf,3);
+		scanCommands(DtoC,AVStr(buf),3);
 	}
-	return bidirectional_relay(Conn,0x2000,0x2000);
+	*/
+	bidirectional_relay(Conn,0x2000,0x2000);
+	return 0;
 }
 
 
@@ -1378,52 +1611,62 @@ static short Trans[][32] = {
  0
 };
 
-static getString(buf,cc,c0,str)
-	unsigned char *buf,*str;
-{	int ci,ch,si;
+static int getString(PCStr(buf),int cc,int c0,PVStr(str))
+{	int ci,si;
+	unsigned char ch;
 
 	si = 0;
 	for(ci = c0; ci < cc; ci++ ){
-		ch = str[si++] = buf[ci];
+		assertVStr(str,str+si+1);
+		ch = buf[ci];
+		setVStrElemInc(str,si,ch); /**/
 		if( ch < ' ' || 0x7F < ch )
 			break;
 	}
-	str[si] = 0;
+	setVStrEnd(str,si);
 	return ci;
 }
-static removeString(buf,cc,from,to)
-	char *buf;
+static int removeString(PVStr(buf),int cc,int from,int to)
 {	int len,ci;
 
 	len = to - from;
 	for( ci = from; ci < cc; ci++ )
-		buf[ci] = buf[ci+len];
+		setVStrElem(buf,ci,buf[ci+len]); /**/
 	return -len;
 }
-static replaceString(buf,cc,from,to,new)
-	char *buf,*new;
+static int replaceString(PVStr(buf),int cc,int from,int to,PCStr(xnew))
 {	int len,nlen,inc;
 	int ci;
 
 	len = to - from;
-	nlen = strlen(new);
+	nlen = strlen(xnew);
 	inc = nlen - len;
 
-	if( 0 < inc )
-		for( ci = cc-1; to <= ci; ci-- ) buf[ci+inc] = buf[ci];
-	else	for( ci = to;   ci <  cc; ci++ ) buf[ci+inc] = buf[ci];
-	for( ci = 0; ci < nlen; ci++ )
-		buf[from+ci] = new[ci];
+	if( 0 < inc ){
+		for( ci = cc-1; to <= ci; ci-- ){
+			assertVStr(buf,buf+ci+inc);
+			setVStrElem(buf,ci+inc,buf[ci]); /**/
+		}
+	}else{
+		for( ci = to;   ci <  cc; ci++ ){
+			assertVStr(buf,buf+ci+inc);
+			setVStrElem(buf,ci+inc,buf[ci]); /**/
+		}
+	}
+	for( ci = 0; ci < nlen; ci++ ){
+		assertVStr(buf,buf+from+ci);
+		setVStrElem(buf,from+ci,xnew[ci]); /**/
+	}
 	return inc;
 }
 
-rewriteTelnet(direction,buf,cc)
-	unsigned char *buf;
+static int rewriteTelnet(int direction,PVStr(buf),int cc)
 {	int si,sji,sjp;
-	int chi,chp;
+	int chp;
+	unsigned char chi;
 	short *sb1;
 	int inc = 0,inc1;
-	char str[4][2048];
+	ACStr(str,4,2048);
 	int strx,start_sjp,start_sji;
 
 	for( si = 0; ; si++ ){
@@ -1452,7 +1695,7 @@ rewriteTelnet(direction,buf,cc)
 			break;
 
 		  case STR:
-			sji = getString(buf,cc,sji,str[strx]);
+			sji = getString(buf,cc,sji,EVStr(str[strx]));
 			strx++;
 			break;
 
@@ -1460,7 +1703,7 @@ rewriteTelnet(direction,buf,cc)
 			if( direction == CtoS )
 			if( Xproxy[0] && !streq(Xproxy,str[0]) ){
 				sv1log("DISPLAY [%s] -> [%s]\n",str[0],Xproxy);
-				inc1 = replaceString(buf,cc,start_sji,sji,
+				inc1 = replaceString(AVStr(buf),cc,start_sji,sji,
 					Xproxy);
 				inc += inc1;
 				sji += inc1;
@@ -1473,7 +1716,7 @@ rewriteTelnet(direction,buf,cc)
 			if( streq(str[0],"DISPLAY") )
 			if( direction == CtoS )
 			if( Xproxy[0] && !streq(Xproxy,str[1]) ){
-				inc1 = replaceString(buf,cc,start_sji,sji,
+				inc1 = replaceString(AVStr(buf),cc,start_sji,sji,
 					Xproxy);
 				inc += inc1;
 				sji += inc1;
