@@ -83,6 +83,10 @@ Description:
 	   - relay among a connected party
 	   - (possiblly with some data conversion)
 
+	PROXY protocol
+	   150 start relaying
+	   - relay in specified protocol (not as circuit level proxy)
+
 	FORWARD [#SID] HR:PR [TRACE[/app-proto]]
 	   260 start forwarding
 	   261 done forwarding
@@ -148,9 +152,16 @@ TODO:
 //////////////////////////////////////////////////////////////////////#*/
 #include <stdio.h>
 #include "delegate.h"
+#include "fpoll.h"
+#include "file.h"
+#include "auth.h"
+
+int service_http(Connection *Conn);
+int service_ftp(Connection *Conn);
 
 #define NEWSOCK		-1,NULL,NULL,NULL
 #define ANYPORT		NULL,0
+#define VStrANYPORT	VStrNULL,0
 
 int acceptedViaVSAP;
 
@@ -175,14 +186,14 @@ int acceptedViaVSAP;
 #define NO_CONNECT	540
 
 #define VER	"VSAP/0.1"
-char *VSAP_VER(){ return VER; }
+const char *VSAP_VER(){ return VER; }
 
-static SockPrintf(sock,fmt,a,b,c,d,e,f,g)
-	char *fmt,*a,*b,*c,*d,*e,*f,*g;
-{	char msg[1024];
+static int SockPrintf(int sock,PCStr(fmt),...)
+{	CStr(msg,1024);
 	int wcc;
+	VARGS(8,fmt);
 
-	sprintf(msg,fmt,a,b,c,d,e,f,g);
+	sprintf(msg,fmt,VA8);
 	wcc = write(sock,msg,strlen(msg));
 	daemonlog("D","I-SAY: %s",msg);
 	return wcc;
@@ -195,29 +206,57 @@ typedef struct {
 	int	a_sock;
 	int	a_id;
 } ASSOCK;
-static ASSOCK SvSocks[32];
-static int SvSockN;
-static ASSOCK ClSocks[32];
-static int ClSockN;
-static int SockID;
-static add_clsock(sock)
+typedef struct {
+	ASSOCK	ve_SvSocks[32]; /**/
+	int	ve_SvSockN;
+	ASSOCK	ve_ClSocks[32]; /**/
+	int	ve_ClSockN;
+	int	ve_SockID;
+	int	ve_bind_exlock;
+	int	ve_bind_nlisten;
+} VsapServEnv;
+static VsapServEnv *vsapServEnv;
+#define SvSocks		vsapServEnv->ve_SvSocks
+#define SvSockN		vsapServEnv->ve_SvSockN
+#define ClSocks		vsapServEnv->ve_ClSocks
+#define ClSockN		vsapServEnv->ve_ClSockN
+#define SockID		vsapServEnv->ve_SockID
+#define bind_exlock	vsapServEnv->ve_bind_exlock
+#define bind_nlisten	vsapServEnv->ve_bind_nlisten
+void minit_vsapsv()
 {
+	if( vsapServEnv == 0 ){
+		vsapServEnv = NewStruct(VsapServEnv);
+		bind_exlock = -1;
+		bind_nlisten = -1;
+	}
+}
+
+static int add_clsock(int sock)
+{
+	if( elnumof(ClSocks) <= ClSockN+1 ){
+		return -1;
+	}
 	ClSockN++;
 	ClSocks[ClSockN].a_id = ++SockID;
 	ClSocks[ClSockN].a_sock = sock;
+	return SockID;
 }
-static del_clsock(sock){
+static void del_clsock(int sock){
 }
-static add_svsock(sock)
+static int add_svsock(int sock)
 {
+	if( elnumof(SvSocks) <= SvSockN+1 ){
+		return -1;
+	}
 	SvSockN++;
 	SvSocks[SvSockN].a_id = ++SockID;
 	SvSocks[SvSockN].a_sock = sock;
 	return SvSockN;
 }
-static del_svsock(sock){
+static void del_svsock(int sock){
 }
-static closeSvSocks()
+static void closeSvSocks()
 {	int fi,svsock;
 
 	for( fi = 1; fi <= SvSockN; fi++ ){
@@ -229,12 +268,12 @@ static closeSvSocks()
 	SvSockN = 0;
 }
 
-static do_connect(lsock,rport,toC)
-	char *rport;
-{	char host[256];
+static int do_connect(int lsock,PCStr(rport),int toC)
+{	CStr(host,256);
 	int port;
 	int clsock;
-	char sockname[256],peername[256];
+	CStr(sockname,256);
+	CStr(peername,256);
 	int wcc;
 	int sockid;
 
@@ -253,16 +292,16 @@ static do_connect(lsock,rport,toC)
 	strcpy(host,"127.0.0.1");
 	port = 0;
 	if( strchr(rport,':') )
-		sscanf(rport,"%[^:]:%d",host,&port);
+		Xsscanf(rport,"%[^:]:%d",AVStr(host),&port);
 	else	sscanf(rport,"%d",&port);
 
-	clsock = Socket1("VSAP",lsock,NULL,NULL,NULL,ANYPORT,host,port,0,NULL,0);
+	clsock = Socket1("VSAP",lsock,NULL,NULL,NULL,VStrANYPORT,host,port,0,NULL,0);
 	if( clsock < 0 ){
 		wcc = SockPrintf(toC,"%s %d cannot connect %s\r\n",
 			VER,NO_CONNECT,rport);
 		return -1;
 	}
-	getpairName(clsock,sockname,peername);
+	getpairName(clsock,AVStr(sockname),AVStr(peername));
 
 CONNECTED:
 	sockid = add_clsock(clsock);
@@ -272,27 +311,22 @@ CONNECTED:
 	return clsock;
 }
 
-static int bind_exlock = -1;
-static int bind_nlisten = -1;
-
-static NewPort(lport,host,port,toC,nlisten)
-	char *lport;
-	char *host;
+static int NewPort(PCStr(lport),PVStr(host),int port,int toC,int nlisten)
 {	int timeout1;
 	int elapsed;
 	int sock;
 	int rcode;
-	char exlockpath[1024];
+	CStr(exlockpath,1024);
 
 	bind_exlock = -1;
 	bind_nlisten = -1;
 
 	if( port == 0 ){
-		sock = Socket1("VSAP",NEWSOCK,host,port,ANYPORT,nlisten,NULL,0);
+		sock = Socket1("VSAP",NEWSOCK,AVStr(host),port,ANYPORT,nlisten,NULL,0);
 		return sock;
 	}
 
-	bind_exlock = PortLocks(lport,1,exlockpath);
+	bind_exlock = PortLocks(lport,1,AVStr(exlockpath));
 	if( bind_exlock < 0 )
 		return -1;
 
@@ -307,7 +341,7 @@ static NewPort(lport,host,port,toC,nlisten)
 	for(;;){
 		rcode = lock_exclusiveTO(bind_exlock,timeout1,&elapsed);
 		if( rcode == 0 ){
-			sock = Socket1("VSAP",NEWSOCK,host,port,ANYPORT,nlisten,NULL,0);
+			sock = Socket1("VSAP",NEWSOCK,AVStr(host),port,ANYPORT,nlisten,NULL,0);
 			break;
 		}
 		if( elapsed < timeout1 )
@@ -317,11 +351,10 @@ static NewPort(lport,host,port,toC,nlisten)
 	}
 	return sock;
 }
-static do_bind(sock,lport,opts,sharedp,sockname,toC)
-	char *lport,*opts,*sockname;
-	int *sharedp;
+static int do_bind(int sock,PVStr(lport),PCStr(opts),int *sharedp,PVStr(sockname),int toC)
 {	int svsock;
-	char host[256],ports[256];
+	CStr(host,256);
+	CStr(ports,256);
 	int port;
 	int nlisten;
 	int wcc;
@@ -334,7 +367,7 @@ static do_bind(sock,lport,opts,sharedp,sockname,toC)
 	nlisten = 0;
 
 	if( lport[0] == '/' ){
-		svsock = server_open_un("vsap",lport,1);
+		svsock = server_open_un("vsap",AVStr(lport),1);
 		daemonlog("E","## SVPORT %s [%d]\n",lport,svsock);
 		if( svsock < 0 ){
 		wcc = SockPrintf(toC,"%s %d not bound-1\r\n",VER,NO_BIND);
@@ -346,8 +379,8 @@ static do_bind(sock,lport,opts,sharedp,sockname,toC)
 	}
 
 	if( strchr(lport,':') )
-		sscanf(lport,"%[^:]:%s",host,ports);
-	else	sscanf(lport,"%s",ports);
+		Xsscanf(lport,"%[^:]:%s",AVStr(host),AVStr(ports));
+	else	Xsscanf(lport,"%s",AVStr(ports));
 	port = atoi(ports);
 
 	if( strncmp(opts,"-l=",3) == 0 )
@@ -363,7 +396,7 @@ static do_bind(sock,lport,opts,sharedp,sockname,toC)
 		daemonlog("D","## RESV_PORT %d\n",svsock);
 		*sharedp = 1;
 	}else
-	if( 0 <= (svsock = NewPort(lport,host,port,toC,nlisten)) ){
+	if( 0 <= (svsock = NewPort(lport,AVStr(host),port,toC,nlisten)) ){
 		daemonlog("D","## NEW_PORT %d\n",svsock);
 		*sharedp = 0;
 	}else{
@@ -371,7 +404,7 @@ static do_bind(sock,lport,opts,sharedp,sockname,toC)
 			lport);
 		return -1;
 	}
-	gethostName(svsock,sockname,"%A:%P");
+	gethostName(svsock,AVStr(sockname),"%A:%P");
 
 BOUND:
 	sockid = add_svsock(svsock);
@@ -380,20 +413,23 @@ BOUND:
 	return svsock;
 }
 
-static do_accept(lport,opts,shared,priobase,fromC,toC)
-	char *lport,*opts;
+static int do_accept(PCStr(lport),PCStr(opts),int shared,int priobase,int fromC,int toC)
 {	int shlock,exlock;
-	char shlockpath[1024],exlockpath[1024];
-	char host[1024];
+	CStr(shlockpath,1024);
+	CStr(exlockpath,1024);
+	CStr(host,1024);
 	int port;
-	char *op,opt1[128];
+	const char *op;
+	CStr(opt1,128);
 	int rcode;
 	int clsock;
 	int wcc;
-	char sockname[256],peername[256];
+	CStr(sockname,256);
+	CStr(peername,256);
 	int fi,fdv[32],rfv[32],nready;
 	int svsock;
 	int priority,timeout;
+	int start;
 
 	clsock = -1;
 	if( SvSockN < 1 ){
@@ -408,7 +444,7 @@ static do_accept(lport,opts,shared,priobase,fromC,toC)
 	host[0] = 0;
 	port = 0;
 	if( strchr(host,':') )
-		sscanf(lport,"%[^:]:%d",host,&port);
+		Xsscanf(lport,"%[^:]:%d",AVStr(host),&port);
 	else	sscanf(lport,"%d",&port);
 
 	priority = 0;
@@ -426,22 +462,39 @@ static do_accept(lport,opts,shared,priobase,fromC,toC)
 	daemonlog("D","accept: priority=%d timeout=%d\n",priority,timeout);
 
 	if( shared ){
-		shlock = PortLocks(lport,2,shlockpath);
+		const char *ltype;
+		int rem;
+		start = time(0L);
+		shlock = PortLocks(lport,2,AVStr(shlockpath));
+		rcode = -1;
 		for(;;){
-			if( priority )
+			if( priority ){
 				rcode = lock_exclusiveTO(shlock,1000,NULL);
-			else	rcode = lock_sharedTO(shlock,1000,NULL);
-			if( !IsConnected(toC,NULL) ){
+				ltype = "exclusive";
+			}else{	rcode = lock_sharedTO(shlock,1000,NULL);
+				ltype = "shared";
+			}
+			if( !IsAlive(toC) ){
 			daemonlog("E","## disconnected during LOCKing\n");
 				goto EXIT;
 			}
 			if( rcode == 0 )
 				break;
+			if( timeout <= time(0L) - start )
+				break;
 		}
+		rem = timeout - (time(0L) - start);
+		if( rcode != 0 && rem <= 0 ){
+		daemonlog("E","## accept: timedout during %s LOCKing (%d)\n",
+				ltype,timeout);
+			goto EXIT;
+		}
+		daemonlog("D","## accept: %s LOCKed (%d)\n",ltype,timeout-rem);
 	}
 
 	daemonlog("D","## START accept at %d (%d ports)\n",port,SvSockN);
 	for(;;){
+	RETRY:
 		fdv[0] = fromC;
 		for( fi = 1; fi <= SvSockN; fi++ ){
 			fdv[fi] = SvSocks[fi].a_sock;
@@ -452,8 +505,11 @@ static do_accept(lport,opts,shared,priobase,fromC,toC)
 		nready = PollIns(timeout*1000,1+SvSockN,fdv,rfv);
 		if( nready <= 0 )
 			break;
-		if( rfv[0] != 0 && !IsConnected(fromC,NULL) )
+		if( rfv[0] != 0 && !IsAlive(fromC) )
+		{
+			daemonlog("E","## disconnected during POLLing\n");
 			break;
+		}
 
 		svsock = -1;
 		for( fi = 1; fi <= SvSockN; fi++ ){
@@ -467,14 +523,41 @@ static do_accept(lport,opts,shared,priobase,fromC,toC)
 		if( svsock < 0 )
 			break;
 
-		/* must lock the socket before ACCEPT on Solaris */
-		exlock = PortLocks(lport,3,exlockpath);
-		clsock = ACCEPT(svsock,1,exlock,0);
+/* not lport but the real port-number of svsock ? */
+		exlock = PortLocks(lport,3,AVStr(exlockpath));
+		start = time(0L);
+		rcode = -1;
+		for(;;){
+			rcode = lock_exclusiveTO(exlock,1000,NULL);
+			if( !IsAlive(toC) ){
+			daemonlog("E","## disconnected during ACCEPT LOCKing\n");
+				rcode = -1;
+				break;
+			}
+			if( rcode == 0 )
+				break;
+			if( timeout <= time(0L) - start ){
+			daemonlog("E","## timedout during ACCEPT LOCKing\n");
+				lock_unlock(exlock);
+				goto RETRY;
+			}
+		}
+		if( rcode == 0 ){
+			if( PollIn(svsock,1) <= 0 ){
+			daemonlog("E","## be snatched during ACCEPT LOCKing\n");
+				lock_unlock(exlock);
+				goto RETRY;
+			}else{
+				clsock = ACCEPT(svsock,1,-1,timeout);
+			}
+		}
+
 		if( !shared )
 			closeSvSocks();
 
 		if( 0 <= exlock )
 			lock_unlock(exlock);
+
 		if( 0 <= shlock ){
 			lock_unlock(shlock);
 			shlock = -1;
@@ -499,11 +582,9 @@ EXIT:
 	return clsock;
 }
 
-static intcom(arg,src,dst,rccp,wccp)
-	char *arg;
-	int *rccp,*wccp;
+static int intcom(PCStr(arg),int src,int dst,int *rccp,int *wccp)
 {	int rcc,wcc,rcode;
-	char req[1024];
+	CStr(req,1024);
 
 	daemonlog("E","## intcom: %d->d\n",src,dst);
 	rcc = RecvLine(src,req,sizeof(req));
@@ -516,14 +597,14 @@ static intcom(arg,src,dst,rccp,wccp)
 		rcode = -1;
 	return rcode;
 }
-static do_forward(Conn,lport,peername,shared,svsock,priority,fromC,toC)
-	Connection *Conn;
-	char *lport,*peername;
+static void do_forward(Connection *Conn,PCStr(lport),PCStr(peername),int shared,int svsock,int priority,int fromC,int toC)
 {	int cls,svs;
-	char host[256];
+	CStr(host,256);
 	int port;
-	char opt[256];
-	char clntname[1024],servname[1024],sockname[1024];
+	CStr(opt,256);
+	CStr(clntname,1024);
+	CStr(servname,1024);
+	CStr(sockname,1024);
 
 	if( svsock < 0 ){
 		SockPrintf(toC,"%s %d no socket to accept\r\n",VER,NO_ACCEPT);
@@ -532,7 +613,7 @@ static do_forward(Conn,lport,peername,shared,svsock,priority,fromC,toC)
 	SockPrintf(toC,"%s %d forwarding start.\r\n",VER,OK_FORWARD1);
 
 	opt[0] = 0;
-	sscanf(peername,"%[^:]:%d %s",host,&port,opt);
+	Xsscanf(peername,"%[^:]:%d %s",AVStr(host),&port,AVStr(opt));
 
 	for(;;){
 		cls = do_accept(lport,"",shared,priority,fromC,toC);
@@ -546,24 +627,24 @@ static do_forward(Conn,lport,peername,shared,svsock,priority,fromC,toC)
 				VER,NO_ACCEPT,lport);
 			break;
 		}
-		getpairName(cls,sockname,clntname);
+		getpairName(cls,AVStr(sockname),AVStr(clntname));
 		SockPrintf(toC,"%s %d accepted %s %s\r\n",VER,OK_ACCEPT,
 			sockname,clntname);
 
-		svs = Socket1("VSAPdata",NEWSOCK,ANYPORT,host,port, 0,NULL,0);
+		svs = Socket1("VSAPdata",NEWSOCK,VStrANYPORT,host,port, 0,NULL,0);
 		if( svs < 0 ){
 			SockPrintf(toC,"%s %d could not connect to %s\r\n",
 				VER,NO_CONNECT,peername);
 			break;
 		}else{
 			if( strstr(opt,"TRACE") ){
-				getpairName(svs,sockname,servname);
+				getpairName(svs,AVStr(sockname),AVStr(servname));
 				SockPrintf(cls,"%s %d %s %s connected.\r\n",
 					VER,OK_CONNECT,sockname,servname);
 				SockPrintf(svs,"%s %d %s %s accepted.\r\n",
 					VER,OK_ACCEPT,sockname,clntname);
 			}
-			relay2_cntl(60*1000,cls,svs,svs,cls,fromC,toC,intcom,0);
+			relay2_cntl(60*1000,cls,svs,svs,cls,fromC,toC,(IFUNCP)intcom,0);
 			close(svs);
 			SockPrintf(toC,"%s %d relay done.\r\n",VER,OK_RELAYED);
 		}
@@ -572,16 +653,14 @@ static do_forward(Conn,lport,peername,shared,svsock,priority,fromC,toC)
 	SockPrintf(toC,"%s %d forwarding done.\r\n",VER,OK_FORWARD2);
 }
 
-static vsap_permit(Conn,portname)
-	Connection *Conn;
-	char *portname;
-{	char host[512];
+static int vsap_permit(Connection *Conn,PCStr(portname))
+{	CStr(host,512);
 	int port;
 
 	host[0] = 0;
 	port = 0;
 	if( strchr(portname,':') )
-		sscanf(portname,"%[^:]:%d",host,&port);
+		Xsscanf(portname,"%[^:]:%d",AVStr(host),&port);
 	else{
 		sscanf(portname,"%d",&port);
 	}
@@ -594,23 +673,34 @@ static vsap_permit(Conn,portname)
 }
 
 
-VSAP_isMethod(request)
-	char *request;
+int VSAP_isMethod(PCStr(request))
 {
 	if( strncmp(request,"VSAP/",5) == 0 )
 		return 1;
 	return 0;
 }
-service_vsap(Conn)
-	Connection *Conn;
-{	char request[1024],reqver[128],*req;
-	int svsock,shared,clsock,rcode;
-	char myport[256];
-	char sockname[256],peername[256];
-	int wcc,rcc;
-	char com[1024],arg[1024],*argp;
-	int timeout;
 
+extern int IO_TIMEOUT;
+
+int service_vsap(Connection *Conn)
+{	CStr(request,1024);
+	CStr(reqver,128);
+	const char *req;
+	int svsock,shared,clsock,rcode;
+	CStr(myport,256);
+	CStr(sockname,256);
+	CStr(peername,256);
+	int wcc,rcc;
+	CStr(com,1024);
+	CStr(arg,1024);
+	const char *argp;
+	CStr(opt,32);
+	const char *op;
+	int timeout;
+	int AuthOk;
+	FILE *authout;
+
+	minit_vsapsv();
 	if( ToS <= 0 || FromS <= 0 ){
 		/*
 		If the DST_HOST is not local
@@ -623,10 +713,13 @@ service_vsap(Conn)
 			connect_to_serv(Conn,FromC,ToC,0);
 		relay_svcl(Conn,FromC,ToC,FromS,ToS);
 		close(ToS);
-		return;
+		return 0;
 	}
 
+	/*
 	timeout = 300;
+	*/
+	timeout = IO_TIMEOUT;
 
 	shared = 0;
 	myport[0] = 0;
@@ -636,15 +729,34 @@ service_vsap(Conn)
 	svsock = -1;
 	reqver[0] = 0;
 
+	authout = TMPFILE("VSAP-AUTH");
+	if( doAUTH(Conn,NULL,authout,"vsap","-",0,CVStr("user-xxxx:pass-xxxx"),CVStr("host-xxxx"),NULL,NULL) == EOF ){
+		AuthOk = 0;
+	}else	AuthOk = -1;
+
 	if( ImMaster ){
 		sprintf(myport,"%s:%d",DST_HOST,DST_PORT);
 	}else
 	for(;;){
-		if( DDI_fgetsFromCbuf(Conn,request,sizeof(request)) == 0 )
+		if( DDI_fgetsFromCbuf(Conn,AVStr(request),sizeof(request),NULL) == 0 )
+		{	int closed = 0;
+			for(;;){
+				if( PollIn(FromC,1*1000) != 0 )
+					break;
+				closed |= checkCloseOnTimeout(1);
+				if( 0 <= clsock && !IsAlive(clsock) ){
+daemonlog("E","## disconnected by peer\n");
+SockPrintf(ToC,"%s %d %s.\r\n",VER,NO_GENERIC_BYE,"disconnected by peer");
+					close(clsock);del_clsock(clsock);
+					goto EXIT;
+				}
+			}
 		if( (rcc = RecvLine(FromC,request,sizeof(request))) <= 0 )
 			break;
+		}
 
 		daemonlog("D","CLIENT-SAYS: %s",request);
+daemonlog("E","CLIENT-SAYS: %s",request);
 		req = request;
 		if( strncmp(req,"VSAP/",5) == 0 )
 			req = wordScan(req,reqver);
@@ -653,11 +765,25 @@ service_vsap(Conn)
 		arg[0] = 0;
 		lineScan(argp,arg);
 
+		if( strcasecmp(com,"AUTH") == 0 ){
+			CStr(ahost,256);
+			ahost[0] = 0;
+			if( doAUTH(Conn,NULL,authout,"vsap","-",0,AVStr(arg),AVStr(ahost),NULL,NULL) == EOF ){
+			}else{
+				AuthOk = 1;
+				SockPrintf(ToC,"%s %d OK\r\n",VER,OK_GENERIC);
+				continue;
+			}
+		}
+		if( AuthOk == 0 ){
+			SockPrintf(ToC,"%s %d forbidden\r\n",VER,NO_PERMISSION);
+			sv1log("WITH AUTHORIZER, but NO AUTH from client\n");
+			break;
+		}
+
 		if( strcasecmp(com,"ECHO") == 0 ){
-			char stime[64];
-			extern char *TIMEFORM_HTTPD;
-			StrftimeLocal(stime,sizeof(stime),
-				TIMEFORM_HTTPD,time(0));
+			CStr(stime,64);
+			StrftimeLocal(AVStr(stime),sizeof(stime),TIMEFORM_HTTPD,time(0),0);
 			SockPrintf(ToC,"%s %d [%s] %s\r\n",VER,OK_GENERIC,
 				stime,arg);
 		}else
@@ -670,18 +796,18 @@ service_vsap(Conn)
 				break;
 		}else
 		if( strcasecmp(com,"BIND") == 0 ){
-			char opts[1024];
+			CStr(opts,1024);
 			opts[0] = 0;
-			sscanf(arg,"%s %[^\r\n]",myport,opts);
+			Xsscanf(arg,"%s %[^\r\n]",AVStr(myport),AVStr(opts));
 			if( !vsap_permit(Conn,myport) )
 				break;
-			svsock = do_bind(-1,myport,opts,&shared,sockname,ToC);
+			svsock = do_bind(-1,AVStr(myport),opts,&shared,AVStr(sockname),ToC);
 			if( svsock < 0 )
 				break;
 		}else
 		if( strcasecmp(com,"LISTEN") == 0 ){
 			int nlisten = atoi(arg);
-			Socket1("VSAP",svsock,NULL,NULL,NULL,ANYPORT,ANYPORT,nlisten,NULL,0);
+			Socket1("VSAP",svsock,NULL,NULL,NULL,VStrANYPORT,ANYPORT,nlisten,NULL,0);
 			SockPrintf(ToC,"%s %d listen ok.\r\n",VER,OK_LISTEN);
 		}else
 		if( strcasecmp(com,"ACCEPT") == 0 ){
@@ -703,7 +829,7 @@ service_vsap(Conn)
 			if( myport[0] == '/' ){
 				strcpy(sockname,myport);
 				strcpy(peername,myport);
-			}else	getpairName(clsock,sockname,peername);
+			}else	getpairName(clsock,AVStr(sockname),AVStr(peername));
 			wcc = SockPrintf(ToC,"%s %d %d %s %s accepted.\r\n",
 				VER,OK_ACCEPT,ClSockN,sockname,peername);
 		}else
@@ -715,7 +841,38 @@ service_vsap(Conn)
 			do_forward(Conn,myport,arg,shared,svsock,0,FromC,ToC);
 		}else
 		if( strcasecmp(com,"RELAY") == 0 ){
+/*
+-t timeout
+ */
+			for( op = arg; *op == '-'; ){
+				if( strneq(op,"-t=",3) ){
+					int to;
+					op = numscanX(op+3,AVStr(opt),sizeof(opt));
+					to = atoi(opt);
+					if( to < timeout )
+						timeout = to;
+				}else{
+					break;
+				}
+			}
+
 			tcp_relay2(timeout*1000,FromC,clsock,clsock,ToC);
+			/*
+			set_linger(clsock,10);
+			*/
+			set_linger(clsock,LIN_TIMEOUT);
+			close(clsock);del_clsock(clsock);
+			break;
+		}else
+		if( strcasecmp(com,"PROXY") == 0 ){
+			ToS = FromS = ToC;
+			ToC = FromC = clsock;
+daemonlog("E","##### VSAP switch protocol to '%s'\n",arg);
+			if( strcmp(arg,"http")==0 )
+				service_http(Conn);
+			else
+			if( strcmp(arg,"ftp")==0 )
+				service_ftp(Conn);
 			set_linger(clsock,10);
 			close(clsock);del_clsock(clsock);
 			break;
@@ -728,6 +885,9 @@ service_vsap(Conn)
 			SockPrintf(ToC,"%s %d %s",VER,NO_GENERIC,request);
 		}
 	}
+EXIT:
+	fclose(authout);
+	return 0;
 }
 
 
@@ -737,31 +897,50 @@ service_vsap(Conn)
 #define FOR_CONNECT	1
 #define FOR_ACCEPT	2
 
-static struct {
-	char   *r_host;
+typedef struct {
+  const	char   *r_host;
 	int	r_port;
-	char   *r_hostport;
+  const	char   *r_hostport;
 	char	r_methods; /* CONNECT, ACCEPT */
-	char   *r_hosts; /* remote client/server hosts to be relayed ? */
-} serverV[8];
-static int serverX;
+  const	char   *r_hosts; /* remote client/server hosts to be relayed ? */
+} ServerV;
+typedef struct {
+	ServerV ve_serverV[8];
+	int	ve_serverX;
+  const	char   *ve_Socknames[256]; /**/
+  const	char   *ve_Peernames[256]; /**/
+} VsapClntEnv;
+static VsapClntEnv *vsapClntEnv;
+#define serverV		vsapClntEnv->ve_serverV
+#define serverX		vsapClntEnv->ve_serverX
+#define Socknames	vsapClntEnv->ve_Socknames
+#define Peernames	vsapClntEnv->ve_Peernames
+void minit_vsapcl(){
+	if( vsapClntEnv == 0 )
+		vsapClntEnv = NewStruct(VsapClntEnv);
+}
 
-static vsap1(vsap)
-	char *vsap;
-{	char host[256];
+static scanListFunc vsap1(PCStr(vsap))
+{	CStr(host,256);
 	int port;
-	char methods[256];
+	CStr(methods,256);
 	int ri;
 
+	if( 8 <= serverX ){
+		daemonlog("F","VSAP[%d] %s -- ignored too many conf.\n",
+			serverX,vsap);
+		return -1;
+	}
+
 	host[0] = 0;
 	port = 0;
 	methods[0] = 0;
-	sscanf(vsap,"%[^:/]%*[:/]%d%*[/:]%s",host,&port,methods);
+	Xsscanf(vsap,"%[^:/]%*[:/]%d%*[/:]%s",AVStr(host),&port,AVStr(methods));
 	daemonlog("E","VSAP[%d] %s:%d\n",serverX,host,port);
 
-	serverV[serverX].r_host = strdup(host);
+	serverV[serverX].r_host = stralloc(host);
 	serverV[serverX].r_port = port;
-	serverV[serverX].r_hostport = strdup(vsap);
+	serverV[serverX].r_hostport = stralloc(vsap);
 
 	if( methods[0] == 0 ){
 		serverV[serverX].r_methods = FOR_CONNECT | FOR_ACCEPT; 
@@ -774,20 +953,19 @@ static vsap1(vsap)
 	serverX++;
 	return 0;
 }
-scan_VSAP(Conn,vsaps)
-	Connection *Conn;
-	char *vsaps;
+void scan_VSAP(Connection *Conn,PCStr(vsaps))
 {
-	scan_commaListL(vsaps,0,vsap1);
+	scan_commaListL(vsaps,0,scanListCall vsap1);
 }
-ViaVSAPassociator(sock)
-{	char sockname[1024],peername[1024];
+int ViaVSAPassociator(int sock)
+{	CStr(sockname,1024);
+	CStr(peername,1024);
 	int ri;
 
 	if( 0 < serverX ){
 		if( sock < 0 )
 			return 1;
-		getpairName(sock,sockname,peername);
+		getpairName(sock,AVStr(sockname),AVStr(peername));
 		/*
 		for( ri = 0; ri < serverX; ri++ ){
 			daemonlog("D","ViaVSAPassociator ? [%s]==[%s,%s]\n",
@@ -799,21 +977,20 @@ ViaVSAPassociator(sock)
 	return 0;
 }
 
-static resp_OK(resp)
-	char *resp;
-{	char scode[128];
+static int resp_OK(PCStr(resp))
+{	CStr(scode,128);
 
 	if( strncmp(resp,VER,sizeof(VER)-1) == 0 )
-	if( 0 < sscanf(resp,"%*s %s",scode) )
+	if( 0 < Xsscanf(resp,"%*s %s",AVStr(scode)) )
 	if( scode[0] == '2' )
 		return atoi(scode);
 	return 0;
 }
 
-static open_vsap(method)
+static int open_vsap(Connection *Conn,int method)
 {	int rsock;
 	int ri;
-	char *host;
+	const char *host;
 	int port;
 
 	if( serverX == 0 )
@@ -826,47 +1003,63 @@ static open_vsap(method)
 				continue;
 		host = serverV[ri].r_host;
 		port = serverV[ri].r_port;
-		rsock = Socket1("VSAP",NEWSOCK,ANYPORT,host,port,0,NULL,0);
+		rsock = Socket1("VSAP",NEWSOCK,VStrANYPORT,host,port,0,NULL,0);
 		if( 0 <= rsock )
 			break;
 	}
+	if( 0 <= rsock ){
+		const char *auth;
+		CStr(authb,256);
+		if( get_MYAUTH(Conn,AVStr(authb),"vsap",host,port) )
+			auth = authb;
+		else	auth = getenv("VSAP_AUTH");
+		if( auth ){
+			CStr(resp,256);
+			SockPrintf(rsock,"%s AUTH %s\r\n",VER,auth);
+			RecvLine(rsock,resp,sizeof(resp));
+			daemonlog((char*)(resp_OK(resp)?"D":"E"),"## AUTH: %s",resp);
+		}
+	}
 	return rsock;
 }
 
-static char *socknames[128];
-static char *peernames[128];
-
-VSAPgetsockname(rsock,sockname)
-	char *sockname;
+int VSAPgetsockname(int rsock,PVStr(sockname))
 {
-	sockname[0] = 0;
-	if( 0 <= rsock && socknames[rsock] != NULL ){
-		strcpy(sockname,socknames[rsock]);
+	setVStrEnd(sockname,0);
+	if( elnumof(Socknames) <= rsock )
+		return -1;
+	if( 0 <= rsock && Socknames[rsock] != NULL ){
+		strcpy(sockname,Socknames[rsock]);
 		return 0;
 	}
 	return -1;
 }
-VSAPgetpeername(rsock,peername)
-	char *peername;
+int VSAPgetpeername(int rsock,PVStr(peername))
 {
-	peername[0] = 0;
-	if( 0 <= rsock && peernames[rsock] != NULL ){
-		strcpy(peername,peernames[rsock]);
+	setVStrEnd(peername,0);
+	if( elnumof(Peernames) <= rsock )
+		return -1;
+	if( 0 <= rsock && Peernames[rsock] != NULL ){
+		strcpy(peername,Peernames[rsock]);
 		return 0;
 	}
 	return -1;
 }
 
-VSAPconnect(sockname,peername)
-	char *sockname,*peername;
-{	char resp[1024];
+int CTX_VSAPconnect(Connection *Conn,PVStr(sockname),PVStr(peername))
+{	CStr(resp,1024);
 	int rsock;
 	int cid;
 	int acode;
 
-	rsock = open_vsap(FOR_CONNECT);
+	rsock = open_vsap(Conn,FOR_CONNECT);
 	if( rsock < 0 )
 		return -1;
+	if( elnumof(Socknames) <= rsock ){
+		close(rsock);
+		return -1;
+	}
+
 
 	if( sockname[0] != 0 ){
 		SockPrintf(rsock,"%s BIND %s\r\n",VER,sockname);
@@ -885,21 +1078,19 @@ VSAPconnect(sockname,peername)
 	}
 
 	SockPrintf(rsock,"RELAY\r\n");
-	sscanf(resp,"%*s %*d %d %s %s connected.",&cid,sockname,peername);
+	Xsscanf(resp,"%*s %*d %d %s %s connected.",&cid,AVStr(sockname),AVStr(peername));
 
-	socknames[rsock] = strdup(sockname);
-	peernames[rsock] = strdup(peername);
+	Socknames[rsock] = stralloc(sockname);
+	Peernames[rsock] = stralloc(peername);
 	return rsock;
 }
-VSAPbind(sockname,nlisten)
-	char *sockname;
-{	Connection ConnBuf,*Conn = &ConnBuf;
-	int rsock;
-	char resp[1024];
+int CTX_VSAPbind(Connection *Conn,PVStr(sockname),int nlisten)
+{	int rsock;
+	CStr(resp,1024);
 	int cid;
 	int acode;
 
-	rsock = open_vsap(FOR_ACCEPT);
+	rsock = open_vsap(Conn,FOR_ACCEPT);
 	if( rsock < 0 )
 		return -1;
 
@@ -908,28 +1099,33 @@ VSAPbind(sockname,nlisten)
 		close(rsock);
 		return -1;
 	}
-	daemonlog("D","VSAPd says: %s",resp);
 	if( !resp_OK(resp) ){
+		daemonlog("E","VSAPd says: %s",resp);
 		close(rsock);
 		return -1;
 	}
+	daemonlog("D","VSAPd says: %s",resp);
 
-	sscanf(resp,"%*s %*d %d %s",&cid,sockname);
+	Xsscanf(resp,"%*s %*d %d %s",&cid,AVStr(sockname));
 	return rsock;
 }
-VSAPaccept(timeout,rsock,priority,sockname,peername)
-	char *sockname,*peername;
-{	char resp[1024];
+int VSAPaccept(int timeout,int rsock,int priority,PVStr(sockname),PVStr(peername))
+{	CStr(resp,1024);
 	int cid;
 	int acode;
 
-	*sockname = *peername = 0;
+	setVStrEnd(peername,0);
+	setVStrEnd(sockname,0);
 
 	if( serverX == 0 )
 		return -1;
 	if( rsock < 0 )
 		return -1;
 
+	if( elnumof(Socknames) <= rsock ){
+		return -1;
+	}
+
 	SockPrintf(rsock,"ACCEPT -p=%d -t=%d\r\n",priority,timeout);
 	if( RecvLine(rsock,resp,sizeof(resp)) <= 0 ){
 		daemonlog("E","## accept failed: VSAPd closed.\n");
@@ -938,7 +1134,18 @@ VSAPaccept(timeout,rsock,priority,sockname,peername)
 
 	daemonlog("D","VSAPd says: %s",resp);
 	if( resp_OK(resp) ){
-		sscanf(resp,"%*s %*d %d %s %s",&cid,sockname,peername);
+		Xsscanf(resp,"%*s %*d %d %s %s",&cid,AVStr(sockname),AVStr(peername));
+/*
+		if( streq(iSERVER_PROTO,"ftp")
+		 || streq(iSERVER_PROTO,"http")
+		){
+daemonlog("E","############ PROXY http\n");
+			SockPrintf(rsock,"PROXY %s\r\n",iSERVER_PROTO);
+		}else
+*/
+/*
+-t timeout
+*/
 		SockPrintf(rsock,"RELAY\r\n");
 	}
 
@@ -948,22 +1155,22 @@ VSAPaccept(timeout,rsock,priority,sockname,peername)
 		return -1;
 	}
 	acceptedViaVSAP = 1;
-	socknames[rsock] = strdup(sockname);
-	peernames[rsock] = strdup(peername);
+	Socknames[rsock] = stralloc(sockname);
+	Peernames[rsock] = stralloc(peername);
 	return 0;
 }
-VSAPbindaccept(timeout,priority,sockname,peername)
-	char *sockname,*peername;
-{	char rport[256];
+int CTX_VSAPbindaccept(Connection *Conn,int timeout,int priority,PVStr(sockname),PVStr(peername))
+{	CStr(rport,256);
 	int rsock,rcode;
 
-	*sockname = *peername = 0;
+	setVStrEnd(peername,0);
+	setVStrEnd(sockname,0);
 
-	printPrimaryPort(rport);
-	if( (rsock = VSAPbind(rport,1)) < 0 )
+	printPrimaryPort(AVStr(rport));
+	if( (rsock = CTX_VSAPbind(Conn,AVStr(rport),1)) < 0 )
 		return -1;
 
-	if( VSAPaccept(timeout,rsock,priority,sockname,peername) < 0 ){
+	if( VSAPaccept(timeout,rsock,priority,AVStr(sockname),AVStr(peername)) < 0 ){
 		close(rsock);
 		rsock = -1;
 	}
